from typing import Optional, Union, Any, Dict, List, Literal, MutableMapping
from collections import OrderedDict
from dataclasses import asdict, dataclass
import pandas as pd
import numpy as np
from scipy.special import exp10

@dataclass
class StatVar:
    """named tuple for value and percentage"""
    
    value: float = 0.0  # value
    perc: float = 0.0  # percentage

@dataclass
class QuantileStat:
    """named tuple for quantile stats"""

    minimum: float
    perc_1th: float
    perc_5th: float
    q1: float
    median: float
    q3: float
    perc_95th: float
    perc_99th: float
    maximum: float
    iqr: float
    range: float

@dataclass
class DescStat:
    """named tuple for descriptive stats"""

    mean: float
    var: float
    std: float
    skew: float
    kurt: float
    mad: float
    cv: float
    normality_p: Optional[float] = None
    
@dataclass
class XtremeStat:
    """named tuple for extreme value stats"""

    lbound: float
    rbound: float
    lxtreme_mean: float
    lxtreme_median: float
    rxtreme_mean: float
    rxtreme_median: float
    
'''Base stat statistics'''
@dataclass    
class BaseStatAttr:
    def serialize(self) -> dict:
        return asdict(self)
    
    @classmethod
    def deserialize(cls, attr: dict):
        return cls(**attr)

@dataclass    
class BaseStatSummary:
    def serialize(self) -> dict:
        raise NotImplementedError("")
    
    @classmethod
    def deserialize(cls, summary: dict):
        raise NotImplementedError("") 
    
@dataclass
class BaseStatObj:
    colname: str
    attr: Any
    summary: Any
    
    def serialize(self) -> dict:
        return dict(
            constructor = self.__class__.__name__,
            colname = self.colname, 
            attr = self.attr.serialize(), 
            summary = self.summary.serialize()
        )
    
    @classmethod
    def deserialize(cls, obj: dict):
        raise NotImplementedError("")
    
    def generate(self, size: int = 100, seed: Optional[int] = None, 
                 ignore_na: bool = False) -> pd.Series:
        """Generate fake data based on learnt distribution

        :param int size: number of samples, defaults to 100
        :param Optional[int] seed: random seed, defaults to None
        :param bool ignore_na: whether to ignore NA value (fill NA with values), defaults to False
        :return pd.Series: generated pandas series
        """
        raise NotImplementedError("")
    
'''CategStat'''

@dataclass
class CategStatAttr(BaseStatAttr):
    pass
    
@dataclass
class CategStatSummary(BaseStatSummary):
    colname_: str
    total_: int
    missing_: StatVar
    unique_: StatVar
    vcounts_: pd.Series
    vpercs_: pd.Series
    
    def serialize(self) -> dict:
        return dict(
            colname=self.colname_,
            stats=dict(
                total=self.total_, 
                missing=asdict(self.missing_), 
                unique=asdict(self.unique_)
            ),
            distribution=dict(
                vcounts=self.vcounts_.to_dict(),
                vpercs=self.vpercs_.to_dict(),
            )
        )
    
    @classmethod
    def deserialize(cls, summary: dict):
        return cls(
            colname_ = summary.get('colname', 'NONAME'),
            total_ = summary["stats"]["total"],
            missing_ = StatVar(**summary["stats"]["missing"]),
            unique_ = StatVar(**summary["stats"]["unique"]),
            vcounts_ = pd.Series(
                summary["distribution"]["vcounts"], 
                name=summary.get('colname', 'NONAME')
            ),
            vpercs_ = pd.Series(
                summary["distribution"]["vpercs"], 
                name=summary.get('colname', 'NONAME')
            ),
        )

@dataclass
class CategStatObj(BaseStatObj):
    colname: str
    attr: CategStatAttr
    summary: CategStatSummary
    
    def generate(self, size: int = 100, seed: Optional[int] = None, 
                 ignore_na: bool = False
    ) -> pd.Series:
        """Generate fake data based on learnt distribution

        :param int size: number of samples, defaults to 100
        :param Optional[int] seed: random seed, defaults to None
        :param bool ignore_na: whether to ignore NA value (fill NA with values), defaults to False
        :return pd.Series: generated pandas series
        """
        rng = np.random.default_rng(seed=seed)  # random generator
        choices = self.summary.vpercs_.index.tolist()
        probs = self.summary.vpercs_.values

        if not ignore_na:
            choices.append(pd.NA)
            probs = list(probs * (1 - self.summary.missing_.perc))
            probs.append(self.summary.missing_.perc)

        s = pd.Series(
            rng.choice(a=choices, size=size, p=probs),
            name=self.colname,
        )
        if self.attr.int_dtype_:
            s = s.astype("float").astype("Int64")
        return s

    
    
'''BinaryStat'''
@dataclass
class BinaryStatAttr(CategStatAttr):
    pos_values_: list
    na_to_pos_: bool = False
    int_dtype_: bool = False
    
@dataclass
class BinaryStatSummary(CategStatSummary):
    binary_vcounts_: pd.Series
    binary_vpercs_: pd.Series
    
    def serialize(self) -> dict:
        return dict(
            colname=self.colname_,
            stats=dict(
                total=self.total_, 
                missing=asdict(self.missing_), 
                unique=asdict(self.unique_)
            ),
            distribution=dict(
                vcounts=self.vcounts_.to_dict(),
                vpercs=self.vpercs_.to_dict(),
            ),
            binary_distribution=dict(
                binary_vcounts=self.binary_vcounts_.to_dict(),
                binary_vpercs=self.binary_vpercs_.to_dict(),
            ),
        )
    
    @classmethod
    def deserialize(cls, summary: dict):
        return cls(
            colname_ = summary.get('colname', 'NONAME'),
            total_ = summary["stats"]["total"],
            missing_ = StatVar(**summary["stats"]["missing"]),
            unique_ = StatVar(**summary["stats"]["unique"]),
            vcounts_ = pd.Series(
                summary["distribution"]["vcounts"], 
                name=summary.get('colname', 'NONAME')
            ),
            vpercs_ = pd.Series(
                summary["distribution"]["vpercs"], 
                name=summary.get('colname', 'NONAME')
            ),
            binary_vcounts_ = pd.Series(
                summary["binary_distribution"]["binary_vcounts"], 
                name=summary.get('colname', 'NONAME')
            ),
            binary_vpercs_ = pd.Series(
                summary["binary_distribution"]["binary_vpercs"], 
                name=summary.get('colname', 'NONAME')
            )
        )
          
    
@dataclass
class BinaryStatObj(CategStatObj):
    colname: str
    attr: BinaryStatAttr
    summary: BinaryStatSummary
    
    @classmethod
    def deserialize(cls, obj: dict):
        return cls(
            colname = obj.get('colname'),
            attr = BinaryStatAttr.deserialize(obj.get('attr')),
            summary = BinaryStatSummary.deserialize(obj.get('summary'))
        )

'''OrdinalCategStat'''
@dataclass
class OrdinalCategStatAttr(CategStatAttr):
    categories_: List[Union[str, int]]
    int_dtype_: bool = False
    
@dataclass
class OrdinalCategStatSummary(CategStatSummary):
    pass
    
@dataclass
class OrdinalCategStatObj(CategStatObj):
    colname: str
    attr: OrdinalCategStatAttr
    summary: OrdinalCategStatSummary
    
    @classmethod
    def deserialize(cls, obj: dict):
        return cls(
            colname = obj.get('colname'),
            attr = OrdinalCategStatAttr.deserialize(obj.get('attr')),
            summary = OrdinalCategStatSummary.deserialize(obj.get('summary'))
        )
    
    
'''NominalCategStat'''
@dataclass
class NominalCategStatAttr(CategStatAttr):
    int_dtype_: bool = False
    max_categories_: int = 25
    
@dataclass
class NominalCategStatSummary(CategStatSummary):
    pass

@dataclass
class NominalCategStatObj(CategStatObj):
    colname: str
    attr: NominalCategStatAttr
    summary: NominalCategStatSummary
    
    @classmethod
    def deserialize(cls, obj: dict):
        return cls(
            colname = obj.get('colname'),
            attr = NominalCategStatAttr.deserialize(obj.get('attr')),
            summary = NominalCategStatSummary.deserialize(obj.get('summary'))
        )

'''NumericStat'''
def safe_asdict(v) -> Optional[dict]:
    if v is None:
        return
    return asdict(v)

def safe_tolist(v: Optional[np.ndarray]) -> Optional[list]:
    if v is None:
        return
    return v.tolist()

def safe_toarray(v: Optional[list]) -> Optional[np.ndarray]:
    if v is None:
        return
    return np.array(v)

def safe_to_stv(params: Optional[dict]) -> Optional[StatVar]:
    if params is None:
        return
    return StatVar(**params)

def safe_to_qst(params: Optional[dict]) -> Optional[QuantileStat]:
    if params is None:
        return
    return QuantileStat(**params)

def safe_to_dst(params: Optional[dict]) -> Optional[DescStat]:
    if params is None:
        return
    return DescStat(**params)

def safe_to_xst(params: Optional[dict]) -> Optional[XtremeStat]:
    if params is None:
        return
    return XtremeStat(**params)

def log10pc(x: np.ndarray) -> np.ndarray:
    """Do log10p transform on both positive and negative range

    :param np.ndarray x: original array
    :return np.ndarray: transformed array
    """
    return np.where(x > 0, np.log10(x + 1), -np.log10(1 - x))

def exp10pc(x: np.ndarray) -> np.ndarray:
    """Do exp10m transform on both positive and negative range

    :param np.ndarray x: original array
    :return np.ndarray: transformed array
    """
    return np.where(x > 0, exp10(x) - 1, -exp10(-x) + 1)
    

@dataclass
class NumericStatAttr(BaseStatAttr):
    setaside_zero_: bool = False
    log_scale_: bool = False
    xtreme_method_: Optional[Literal["iqr", "quantile"]] = None
    bins_: int = 100
    
@dataclass
class NumericStatSummary(BaseStatSummary):
    colname_: str
    total_: int
    missing_: StatVar
    zeros_: StatVar
    infs_pos_: StatVar
    infs_neg_: StatVar
    stat_quantile_: QuantileStat
    stat_descriptive_: DescStat
    hist_: np.ndarray
    bin_edges_: np.ndarray
    xtreme_: StatVar
    xtreme_stat_: Optional[XtremeStat] = None
    xtreme_stat_log_: Optional[XtremeStat] = None
    xtreme_log_: Optional[StatVar] = None
    stat_quantile_log_: Optional[QuantileStat] = None
    stat_descriptive_log_: Optional[DescStat] = None
    hist_log_: Optional[np.ndarray] = None
    bin_edges_log_: Optional[np.ndarray] = None
    
    def serialize(self) -> dict:
        return dict(
            colname=self.colname_,
            stats=dict(
                total=self.total_,
                missing=asdict(self.missing_),
                zeros=asdict(self.zeros_),
                infs_pos=asdict(self.infs_pos_),
                infs_neg=asdict(self.infs_neg_),
            ),
            xtreme_num=dict(
                origin=safe_asdict(self.xtreme_),
                log=safe_asdict(self.xtreme_log_),
            ),  # number of extreme values
            xtreme_stat=dict(
                origin=safe_asdict(self.xtreme_stat_),
                log=safe_asdict(self.xtreme_stat_log_),
            ),  # extreme value statistics
            stat_quantile=dict(
                origin=asdict(self.stat_quantile_),
                log=safe_asdict(self.stat_quantile_log_),
            ),
            stat_descriptive=dict(
                origin=asdict(self.stat_descriptive_),
                log=safe_asdict(self.stat_descriptive_log_),
            ),
            histogram=dict(
                origin=dict(
                    hist=self.hist_.tolist(), 
                    bin_edges=self.bin_edges_.tolist()
                ),
                log=dict(
                    hist=safe_tolist(self.hist_log_),
                    bin_edges=safe_tolist(self.bin_edges_log_)
                ),
            ),
        )
    
    @classmethod
    def deserialize(cls, summary: dict):
        return cls(
            colname_ = summary.get('colname', 'NONAME'),
            total_ = summary["stats"]["total"],
            missing_ = StatVar(**summary["stats"]["missing"]),
            zeros_ = StatVar(**summary["stats"]["zeros"]),
            infs_pos_ = StatVar(**summary["stats"]["infs_pos"]),
            infs_neg_ = StatVar(**summary["stats"]["infs_neg"]),
            xtreme_ = StatVar(**summary["xtreme_num"]["origin"]),
            
            stat_quantile_ = QuantileStat(**summary["stat_quantile"]["origin"]),
            stat_descriptive_ = DescStat(**summary["stat_descriptive"]["origin"]),
            hist_ = np.array(summary["histogram"]["origin"]["hist"]),
            bin_edges_ = np.array(summary["histogram"]["origin"]["bin_edges"]),
            
            # need
            xtreme_stat_ = safe_to_xst(summary["xtreme_stat"]["origin"]),
            xtreme_log_ = safe_to_stv(summary["xtreme_num"]["log"]),
            xtreme_stat_log_ = safe_to_xst(summary["xtreme_stat"]["log"]),
            stat_quantile_log_ = safe_to_qst(summary["stat_quantile"]["log"]),
            stat_descriptive_log_ = safe_to_dst(summary["stat_descriptive"]["log"]),
            hist_log_ = safe_toarray(summary["histogram"]["log"]["hist"]),
            bin_edges_log_ = safe_toarray(summary["histogram"]["log"]["bin_edges"])
        )
   
@dataclass
class NumericStatObj(BaseStatObj):
    colname: str
    attr: NumericStatAttr
    summary: NumericStatSummary
    
    @classmethod
    def deserialize(cls, obj: dict):
        return cls(
            colname = obj.get('colname'),
            attr = NumericStatAttr.deserialize(obj.get('attr')),
            summary = NumericStatSummary.deserialize(obj.get('summary'))
        )
    
    def generate(self, size: int = 100, seed: Optional[int] = None, 
                 ignore_na: bool = False) -> pd.Series:
        """Generate fake data based on learnt distribution

        :param int size: number of samples, defaults to 100
        :param Optional[int] seed: random seed, defaults to None
        :param bool ignore_na: whether to ignore NA value (fill NA with values), defaults to False
        :return pd.Series: generated pandas series
        """
        from scipy.stats import rv_histogram
        
        if ignore_na:
            na_num = 0  # number of na values
            valid_num = size
            na_vs = []
        else:
            na_num = int(size * self.summary.missing_.perc)  # number of na values
            valid_num = size - na_num
            na_vs = [pd.NA] * na_num

        # size of zeros
        if self.attr.setaside_zero_:
            zero_num = int(valid_num * self.summary.zeros_.perc)  # number of zeros
        else:
            zero_num = 0
        zero_vs = np.zeros(shape=zero_num)
        # size of infs
        inf_num = int(valid_num * self.summary.infs_pos_.perc)  # number of +inf
        neg_inf_num = int(valid_num * self.summary.infs_neg_.perc)  # number of -inf
        inf_vs = [np.inf] * inf_num
        neg_inf_vs = [-np.inf] * neg_inf_num

        # remaining valid nums
        valid_num = valid_num - zero_num - inf_num - neg_inf_num

        # valid numerical values
        if self.attr.log_scale_:
            hist_dist = rv_histogram(
                histogram=(self.summary.hist_log_, self.summary.bin_edges_log_), 
                seed=seed
            )
            valid_vs = exp10pc(hist_dist.rvs(size=valid_num, random_state = seed))
        else:
            hist_dist = rv_histogram(
                histogram=(self.summary.hist_, self.summary.bin_edges_), 
                seed=seed
            )
            valid_vs = hist_dist.rvs(size=valid_num, random_state = seed)

        values = np.concatenate((valid_vs, inf_vs, neg_inf_vs, zero_vs, na_vs))
        np.random.seed(seed)
        np.random.shuffle(values)
        np.random.seed(None) # reset

        return pd.Series(
            values,
            name=self.colname,
        ).astype(dtype="Float64")
        

'''Tabular'''

class TabularStat(
        OrderedDict, 
        MutableMapping[str, BaseStatObj]
    ):
    CONSTRUCTORS = {
        NumericStatObj.__name__: NumericStatObj,
        NominalCategStatObj.__name__: NominalCategStatObj,
        OrdinalCategStatObj.__name__: OrdinalCategStatObj,
        BinaryStatObj.__name__: BinaryStatObj,
    }
    """Tabular stat for multiple columns"""
    
    def serialize(self) -> dict:
        return {
            col: obj.serialize() for col, obj in self.items()
        }
    
    @classmethod
    def deserialize(cls, stat_dict: dict):
        r = OrderedDict()
        for col, obj in stat_dict.items():
            c = cls.CONSTRUCTORS.get(obj['constructor']) # constructor
            r[col] = c.deserialize(obj)
        return r

    def get_categ_cols(self) -> list[str]:
        """Get categorical columns (include binary, nominal, ordinal)

        :return list[str]: list of categ columns
        """
        return [
            col
            for col, obj in self.items()
            if isinstance(obj, (BinaryStatObj, NominalCategStatObj, OrdinalCategStatObj))
            or obj.__class__.__name__ in ("BinaryStatObj", "NominalCategStatObj", "OrdinalCategStatObj")
        ]

    def get_nominal_cols(self) -> list[str]:
        """Get nominal columns

        :return list[str]: list of nominal columns
        """
        return [
            col
            for col, obj in self.items()
            if isinstance(obj, NominalCategStatObj)
            or obj.__class__.__name__ == "NominalCategStatObj"
        ]

    def get_binary_cols(self) -> list[str]:
        """Get binary columns

        :return list[str]: list of binary columns
        """
        return [
            col
            for col, obj in self.items()
            if isinstance(obj, BinaryStatObj)
            or obj.__class__.__name__ == "BinaryStatObj"
        ]

    def get_ordinal_cols(self) -> list[str]:
        """Get ordinal columns

        :return list[str]: list of ordinal columns
        """
        return [
            col
            for col, obj in self.items()
            if isinstance(obj, OrdinalCategStatObj)
            or obj.__class__.__name__ == "OrdinalCategStatObj"
        ]

    def get_numeric_cols(self) -> list[str]:
        """Get numeric columns

        :return list[str]: list of numeric columns
        """
        return [
            col
            for col, obj in self.items()
            if isinstance(obj, NumericStatObj)
            or obj.__class__.__name__ == "NumericStatObj"
        ]
        
    def generate(self, size: int = 100, seed: Optional[int] = None, 
                 ignore_na: bool = False
    ) -> pd.DataFrame:
        """Generate fake data based on learnt distribution

        :param int size: number of samples, defaults to 100
        :param Optional[int] seed: random seed, defaults to None
        :param bool ignore_na: whether to ignore NA value (fill NA with values), defaults to False
        :return pd.DataFrame: generated fake data
        """
        return pd.concat(
            [
                obj.generate(
                    size=size, 
                    seed=seed, 
                    ignore_na=ignore_na
                )
                for col, obj in self.items()
            ],
            axis=1,
        )
        

'''Univaraite Corr'''

@dataclass
class BaseUniVarClfTargetCorr:
    colname_: str
    yname_: str
    ylabels_: List[str]
    
@dataclass
class CategUniVarClfTargetCorr(BaseUniVarClfTargetCorr):
    p_x_y_: Dict[str, pd.DataFrame]
    p_y_x_: Dict[str, pd.DataFrame]
    
    def serialize(self) -> dict:
        return dict(
            constructor = self.__class__.__name__,
            colname = self.colname_,
            yname = self.yname_,
            ylables = self.ylabels_,
            p_x_y = {
                k: v.to_dict(orient = 'index')
                for k, v in self.p_x_y_.items() 
            },
            p_y_x = {
                k: v.to_dict(orient = 'index')
                for k, v in self.p_y_x_.items() 
            },
        )

    @classmethod
    def deserialize(cls, stat_dict: dict):
        c = cls()
        c.colname_ = stat_dict['colname']
        c.yname_ = stat_dict['yname']
        c.ylabels_ = stat_dict['ylabels']
        c.p_x_y_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in stat_dict['p_x_y'].items()
        }
        c.p_y_x_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in stat_dict['p_y_x'].items()
        }
        return c

    
@dataclass
class NumericUniVarClfTargetCorr(BaseUniVarClfTargetCorr):
    p_x_y_: Dict[str, pd.DataFrame]
    p_y_x_: pd.DataFrame
    
    def serialize(self) -> dict:
        return dict(
            constructor = self.__class__.__name__,
            colname = self.colname_,
            yname = self.yname_,
            ylables = self.ylabels_,
            p_x_y = {
                k: v.to_dict(orient = 'index')
                for k, v in self.p_x_y_.items() 
            },
            p_y_x = (
                self.p_y_x_
                .replace({np.nan: None})
                .to_dict(orient = 'records')
            ),
        )
    
    @classmethod
    def deserialize(cls, stat_dict: dict):
        c = cls()
        c.colname_ = stat_dict['colname']
        c.yname_ = stat_dict['yname']
        c.ylabels_ = stat_dict['ylabels']
        
        c.p_x_y_ = {
            k: pd.DataFrame.from_dict(v, orient = 'index')
            for k, v in stat_dict['p_x_y'].items()
        }
        c.p_y_x_ = pd.DataFrame.from_records(stat_dict['p_y_x'])
        return c


from pathlib import Path
import pyarrow.parquet as pq
from typing import List, Dict, Union
from fsspec import AbstractFileSystem
import ibis
import pandas as pd

from luntaiDs.CommonTools.dtyper import DSchema
from luntaiDs.CommonTools.utils import save_ibis_to_parquet_on_fs, save_pandas_to_parquet_on_fs
from luntaiDs.CommonTools.warehouse import BaseWarehouseHandler

class _BaseModelDataRegistry:

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        raise NotImplementedError("")

    def register(self, data_id: str, train_ds: Union[ibis.expr.types.Table | pd.DataFrame], 
                 test_ds: Union[ibis.expr.types.Table | pd.DataFrame], replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param Union[ibis.expr.types.Table | pd.DataFrame] train_ds: training dataset
        :param Union[ibis.expr.types.Table | pd.DataFrame] test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        raise NotImplementedError("")

    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        raise NotImplementedError("")

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        raise NotImplementedError("")


class ModelDataRegistryFileSystem(_BaseModelDataRegistry):
    def __init__(self, fs: AbstractFileSystem, data_root: str):
        """modeling data registry implemented for fsspec compatible file system
        file structure:
        
            data_root
            - data_id
                - train.parquet
                - test.parquet

        :param AbstractFileSystem fs: the fsspec compatible filesystem
        :param str data_root: root path, if on object storage, 
            the full path including buckets
        """
        self._fs = fs
        self._data_root = data_root
        self._fs.makedirs(
            path = data_root,
            exist_ok = True
        )

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        try:
            files = self._fs.listdir(path = self._data_root, detail = False)
            ids = [Path(file).name for file in files]
        except FileNotFoundError as e:
            ids = []
        return ids

    def get_train_path(self, data_id: str) -> str:
        return (Path(self._data_root) / data_id / f"train_{data_id}.parquet").as_posix()

    def get_test_path(self, data_id: str) -> str:
        return (Path(self._data_root) / data_id / f"test_{data_id}.parquet").as_posix()

    def register(self, data_id: str, train_ds: Union[ibis.expr.types.Table | pd.DataFrame], 
                 test_ds: Union[ibis.expr.types.Table | pd.DataFrame], replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param Union[ibis.expr.types.Table | pd.DataFrame] train_ds: training dataset
        :param Union[ibis.expr.types.Table | pd.DataFrame] test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        existing_ids = self.get_existing_ids()
        if not replace and data_id in existing_ids:
            raise ValueError(f"data id {data_id} already exist, pls use another id")
        
        self._fs.makedirs(
            path = (Path(self._data_root) / data_id).as_posix(),
            exist_ok = True
        )
        
        if isinstance(train_ds, pd.DataFrame) and isinstance(train_ds, pd.DataFrame):
            save_pandas_to_parquet_on_fs(
                df = train_ds,
                fs = self._fs,
                filepath = self.get_train_path(data_id)
            )
            save_ibis_to_parquet_on_fs(
                df = test_ds,
                fs = self._fs,
                filepath = self.get_test_path(data_id)
            )
        elif isinstance(train_ds, ibis.expr.types.Table) and isinstance(train_ds, ibis.expr.types.Table):
            save_ibis_to_parquet_on_fs(
                df = train_ds,
                fs = self._fs,
                filepath = self.get_train_path(data_id)
            )
            save_ibis_to_parquet_on_fs(
                df = test_ds,
                fs = self._fs,
                filepath = self.get_test_path(data_id)
            )
        else:
            raise TypeError(f"train_ds/test_ds must be either ibis or pandas dataframe")


    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        assert data_id in self.get_existing_ids(), f"data id {data_id} does not exist"
        
        train_filepath = self.get_train_path(data_id)
        test_filepath = self.get_test_path(data_id)
        
        try:
            # attempt to read from duckdb directly if possible
            con = ibis.duckdb.connect()
            con.register_filesystem(self._fs)
            train_ds = ibis.read_parquet(
                self._fs.unstrip_protocol(
                    name = train_filepath
                )
            )
            test_ds = ibis.read_parquet(
                self._fs.unstrip_protocol(
                    name = test_filepath
                )
            )
        except Exception as e:
            # will use fs native open method to read
            with self._fs.open(path = train_filepath) as obj:
                train_arr = pq.read_table(
                    source = obj
                )
                train_ds = ibis.memtable(train_arr)
            with self._fs.open(path = test_filepath) as obj:
                test_arr = pq.read_table(
                    source = obj
                )
                test_ds = ibis.memtable(test_arr)
            
        if target_col:
            X_train = train_ds.drop(target_col)
            X_test = test_ds.drop(target_col)
            y_train = train_ds[target_col]
            y_test = test_ds[target_col]
            return X_train, y_train, X_test, y_test
        else:
            return train_ds, test_ds

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        dpath = (Path(self._data_root) / data_id).as_posix()
        self._fs.rm(
            path = dpath,
            recursive = True
        )


class _ModelDataRegistryWarehouse(_BaseModelDataRegistry):
    DATA_ID_COL = "DATA_ID_"
    TRAIN_TEST_IND_COL = "IS_TRAIN_"
    
    def __init__(self, handler: BaseWarehouseHandler, schema: str, table: str):
        """initialize

        :param BaseWarehouseHandler handler: base warehouse handler
        :param str schema: schema for the table storing modeling data
        :param str table: table storing modeling data
        """
        self.handler = handler
        self.schema = schema
        self.table = table
        
    def init(self, col_schemas: DSchema, **kws):
        """warehouse need create and initialize schema/table structure at beginning

        :param DSchema col_schemas: column schema object
        """
        # create the schema first
        self.handler.create_schema(schema = self.schema)
        
        # validate if table already exists
        if self.handler.is_exist(schema = self.schema, table = self.table):
            # if exists, check if schema changes
            existing_schema = self.handler.get_dtypes(
                schema = self.schema, 
                table = self.table
            )
            current_schema = col_schemas.ibis_schema
            if not current_schema.equals(existing_schema):
                msg = f"""
                {self.schema}.{self.table} table schema changed:
                Current in DB:
                {existing_schema}
                Now:
                {current_schema}
                
                Recreate table schema or fix it
                """
                raise TypeError(msg)
            return
        
        # create the table
        self.handler.create_table(
            schema = self.schema,
            table = self.table,
            col_schemas = col_schemas,
            **kws
        )
    
    def get_table(self) -> ibis.expr.types.Table:
        """get underlying training table using ibis

        :return ibis.expr.types.Table: ibis table
        """
        return self.handler.get_table(schema = self.schema, table = self.table)

    def get_existing_ids(self) -> List[str]:
        """get list of registered data ids

        :return List[str]: data ids
        """
        return (
            self.get_table()
            .select(self.DATA_ID_COL)
            .distinct()
            .to_pandas()
            [self.DATA_ID_COL]
            .tolist()
        )
            

    def fetch(self, data_id: str, target_col: str = None):
        """fetch training/testing dataset

        :param str data_id: data id to be fetched
        :param str target_col: the target column, defaults to None.
        :return:
            - if target_col given, will split to [X_train, y_train, X_test, y_test]
            - if target_col not given, will just split to [train_ds, test_ds]
        """
        table = self.get_table()
        train_ds = (
            table
            .filter((table[self.DATA_ID_COL] == data_id) & (table[self.TRAIN_TEST_IND_COL] == True))
            .drop(self.DATA_ID_COL, self.TRAIN_TEST_IND_COL)
        )
        test_ds = (
            table
            .filter((table[self.DATA_ID_COL] == data_id) & (table[self.TRAIN_TEST_IND_COL] == False))
            .drop(self.DATA_ID_COL, self.TRAIN_TEST_IND_COL)
        )
        if target_col:
            X_train = train_ds.drop(target_col)
            X_test = test_ds.drop(target_col)
            y_train = train_ds[target_col]
            y_test = test_ds[target_col]
            return X_train, y_train, X_test, y_test
        else:
            return train_ds, test_ds


from typing import List, Dict, Union
import ibis
import pandas as pd
import logging
from luntaiDs.ModelingTools.Serving.data_registry import _ModelDataRegistryWarehouse
from luntaiDs.ProviderTools.gcp.dbapi import WarehouseHandlerBQSQL


class _BaseModelDataRegistryBQ(_ModelDataRegistryWarehouse):
    DATA_ID_COL = "DATA_ID_"
    TRAIN_TEST_IND_COL = "IS_TRAIN_"
    
    def __init__(self, handler: WarehouseHandlerBQSQL, schema: str, table: str):
        """initialize

        :param BaseWarehouseHandler handler: base warehouse handler
        :param str schema: schema for the table storing modeling data
        :param str table: table storing modeling data
        """
        super().__init__(
            handler=handler,
            schema=schema,
            table=table
        )
        

    def register(self, data_id: str, train_ds: Union[ibis.expr.types.Table | pd.DataFrame], 
                 test_ds: Union[ibis.expr.types.Table | pd.DataFrame], replace: bool = False):
        """register table in ibis format

        :param str data_id: data id
        :param Union[ibis.expr.types.Table | pd.DataFrame] train_ds: training dataset
        :param Union[ibis.expr.types.Table | pd.DataFrame] test_ds: testing dataset
        :param bool replace: whether to replace existing dataset, defaults to False
        """
        if replace:
            self.remove(data_id = data_id)
            
        if isinstance(train_ds, pd.DataFrame) and isinstance(test_ds, pd.DataFrame):
            # add data_id column
            train_ds.loc[:, self.DATA_ID_COL] = data_id
            test_ds.loc[:, self.DATA_ID_COL] = data_id
            # add train test indicator
            train_ds.loc[:, self.TRAIN_TEST_IND_COL] = True
            test_ds.loc[:, self.TRAIN_TEST_IND_COL] = False
            # save to database
            self.handler.save_pandas(
                df = train_ds,
                schema = self.schema,
                table = self.table
            )
            self.handler.save_pandas(
                df = test_ds,
                schema = self.schema,
                table = self.table
            )
        elif isinstance(train_ds, ibis.expr.types.Table) and isinstance(test_ds, ibis.expr.types.Table):
            # add data_id column
            train_ds = (
                train_ds
                .mutate(ibis.literal(data_id, 'String').name(self.DATA_ID_COL))
                .mutate(ibis.literal(True, 'Boolean').name(self.TRAIN_TEST_IND_COL))
            )
            test_ds = (
                train_ds
                .mutate(ibis.literal(data_id, 'String').name(self.DATA_ID_COL))
                .mutate(ibis.literal(False, 'Boolean').name(self.TRAIN_TEST_IND_COL))
            )
            # save training set to database
            query_train = ibis.to_sql(train_ds)
            qry_cols_train = self.handler.query(query_train).columns
            sql_train = f"""
            INSERT {self.schema}.{self.table} ({','.join(qry_cols_train)})
            {query_train}
            """
            logging.info(f"Inserting into {self.schema}.{self.table} using query:\n{sql_train}")
            self.handler.execute(sql_train)
            # save testing set to database
            query_test = ibis.to_sql(test_ds)
            qry_cols_test = self.handler.query(query_test).columns
            sql_test = f"""
            INSERT {self.schema}.{self.table} ({','.join(qry_cols_test)})
            {query_test}
            """
            logging.info(f"Inserting into {self.schema}.{self.table} using query:\n{sql_test}")
            self.handler.execute(query_test)    
            
        else:
            raise ValueError("train_ds and test_ds must be of type either pandas or ibis dataframe")

    def remove(self, data_id: str):
        """remove dataset from registry

        :param str data_id: data id to be removed
        """
        sql = f"""
        DELETE {self.schema}.{self.table}
        WHERE {self.DATA_ID_COL} = '{data_id}'
        """
        logging.info(f"Deleting table {self.schema}.{self.table} using query:\n{sql}")
        self.handler.execute(sql)


from __future__ import annotations
from datetime import date
from typing import List
import pyarrow as pa
import pyarrow.parquet as pq
import ibis
from luntaiDs.CommonTools.SnapStructure.structure import SnapshotDataManagerFileSystem
from luntaiDs.ProviderTools.ssh.sftp import SFTPFileSystem


class SnapshotDataManagerSFTP(SnapshotDataManagerFileSystem):
    """files are saved as parquet snapshots under each schema.table
        file naming convention: dir/tablename_YYYY-MM-DD.parquet
        using fsspec implementation of SFTP system
    """
    @classmethod
    def setup(cls, fs: SFTPFileSystem, root_dir: str):        
        super(SnapshotDataManagerSFTP, cls).setup(
            fs = fs,
            root_dir = root_dir
        )
        
    def read(self, snap_dt: date, **kws) -> ibis.expr.types.Table:
        """Read as ibis dataframe (one snapshot date) data

        :param snap_dt: snap_dt to load
        :return:
        """
        filepath = self.get_file_path(snap_dt=snap_dt)
        
        with self._fs.open(filepath) as obj:
            df_arr: pa.Table = pq.read_table(obj, **kws)
        return self._ibis_con.read_in_memory(df_arr)
        
    
    def reads(self, snap_dts: List[date], **kws) -> ibis.expr.types.Table:
        """reads as ibis dataframe (vertically concat of several given snap dates data)

        :param snap_dts: list of snap dates to read
        :return:
        """
        filepaths = (
            self.get_file_path(snap_dt) for snap_dt in snap_dts
        )
        def df_iter():
            for filepath in filepaths:
                with self._fs.open(filepath) as obj:
                    df_arr: pa.Table = pq.read_table(obj, **kws)
                    yield df_arr
        
        df_arr = pa.concat_tables(df_iter())
        
        return self._ibis_con.read_in_memory(df_arr)
        


from enum import Enum
from typing import Dict, Literal
import logging
import numpy as np
import pandas as pd
from scipy.stats import ks_2samp
from sklearn.preprocessing import label_binarize
from sklearn.metrics import accuracy_score, balanced_accuracy_score, top_k_accuracy_score, \
        average_precision_score, f1_score, log_loss, precision_score, recall_score, \
        roc_auc_score, roc_curve, precision_recall_curve
from sklearn.metrics._ranking import _binary_clf_curve

def ks_stat(y, yhat):
    return ks_2samp(yhat[y==1], yhat[y!=1]).statistic

def binary_clf_metric_at_thresholds(y_true: np.ndarray, pred_prob: np.ndarray) -> pd.DataFrame:
    """Binary classification metrics at each threshold

    :param y_true: ground truth target array (binary), 1-d array
    :param pred_prob: predicted probabilities array (float between 0-1), 1-d array
    :return: DataFrame of columns [threshold, tp, fp, tn, fn, tpr, fpr, precision, recall]
    """
    fps, tps, thresholds = _binary_clf_curve(y_true, pred_prob)
    num_p = y_true.sum()
    num_n = (1 - y_true).sum()

    t = pd.DataFrame({
        'threshold': thresholds,
        'tp': tps,
        'fp': fps,
        'tn': num_n - fps,
        'fn': num_p - tps
    })
    t['tpr'] = t['tp'] / (t['tp'] + t['fn'])
    t['fpr'] = t['fp'] / (t['fp'] + t['tn'])
    t['precision'] = t['tp'] / (t['tp'] + t['fp'])
    t['recall'] = t['tpr']
    return t



class _SKMetricsMultiCls:
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        """a standardized sklearn scorer, support for multi-class

        :param np.ndarray y_true: ground truth, 1-d array, size (n,)
        :param np.ndarray y_pred_probs: 2-d array, size (n, n_class), the predicted probabilities
            note that the order of the columns should be exactly same as labels in ground truth array
        :return float: the metric result
        """
        raise NotImplementedError("")
    
    def _convert_when_binary(self, y_pred_probs: np.ndarray) -> np.ndarray:
        """sklearn metrics are not behaving same for binary vs. multi-class

        :param np.ndarray y_pred_probs: predicted probabilities, size (n, n_class)
        :return np.ndarray: if binary case, extracted probs for class 1, size (n, )
                            if multi-class case, do nothing, size (n, n_class)
        """
        if y_pred_probs.shape[1] == 2:
            if isinstance(y_pred_probs, pd.DataFrame):
                return y_pred_probs.iloc[:, 1]
            return y_pred_probs[:, 1]
        return y_pred_probs
        

class Accurary(_SKMetricsMultiCls):
    def __init__(self, balanced: bool = False, top_k: int = 1):
        if balanced and top_k != 1:
            logging.warning('if set to balanced, will not consider top_k param')
        self._balanced = balanced
        self._topk = top_k
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        if self._balanced:
            y_pred_labels = np.argmax(y_pred_probs, axis = 1)
            return balanced_accuracy_score(y_true, y_pred_labels)
        if self._topk == 1:
            y_pred_labels = np.argmax(y_pred_probs, axis = 1)
            return accuracy_score(y_true, y_pred_labels)
        if self._topk != 1:
            y_pred_probs = self._convert_when_binary(y_pred_probs)
            return top_k_accuracy_score(y_true, y_pred_probs, k = self._topk)
    
class AvgPrecision(_SKMetricsMultiCls):
    def __init__(self, average: Literal['micro', 'macro', 'weighted']):
        self._average = average
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_true_binary = label_binarize(y_true, classes = list(range(y_pred_probs.shape[1])))
        y_pred_probs = self._convert_when_binary(y_pred_probs)
        return average_precision_score(y_true_binary, y_pred_probs, average = self._average)
    
class F1Score(_SKMetricsMultiCls):
    def __init__(self, average: Literal['micro', 'macro', 'weighted']):
        self._average = average
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_pred_labels = np.argmax(y_pred_probs, axis = 1)
        return f1_score(y_true, y_pred_labels, average = self._average)
    
class Precision(_SKMetricsMultiCls):
    def __init__(self, average: Literal['micro', 'macro', 'weighted']):
        self._average = average
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_pred_labels = np.argmax(y_pred_probs, axis = 1)
        return precision_score(y_true, y_pred_labels, average = self._average)
    
class Recall(_SKMetricsMultiCls):
    def __init__(self, average: Literal['micro', 'macro', 'weighted']):
        self._average = average
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_pred_labels = np.argmax(y_pred_probs, axis = 1)
        return recall_score(y_true, y_pred_labels, average = self._average)
    
class RocAuc(_SKMetricsMultiCls):
    def __init__(self, average: Literal['micro', 'macro', 'weighted']):
        self._average = average
        
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_true_binary = label_binarize(y_true, classes = list(range(y_pred_probs.shape[1])))
        y_pred_probs = self._convert_when_binary(y_pred_probs)
        return roc_auc_score(y_true_binary, y_pred_probs, 
                            average = self._average)
    
class EntropyLoss(_SKMetricsMultiCls):
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray) -> float:
        y_pred_probs = self._convert_when_binary(y_pred_probs)
        return log_loss(y_true, y_pred_probs)


class SKClfMetricsEnum(Enum):
    accuracy = Accurary(balanced=False)
    balanced_accuracy = Accurary(balanced=True)
    top_2_accuracy = Accurary(top_k=2)
    top_5_accuracy = Accurary(top_k=5)
    average_precision_micro = AvgPrecision(average='micro')
    average_precision_macro = AvgPrecision(average='macro')
    average_precision_weighted = AvgPrecision(average='weighted')
    f1_micro = F1Score(average='micro')
    f1_macro = F1Score(average='macro')
    f1_weighted = F1Score(average='weighted')
    precision_micro = Precision(average='micro')
    precision_macro = Precision(average='macro')
    precision_weighted = Precision(average='weighted')
    recall_micro = Recall(average='weighted')
    recall_macro = Recall(average='macro')
    recall_weighted = Recall(average='weighted')
    roc_auc_micro = RocAuc(average='micro')
    roc_auc_macro = RocAuc(average='macro')
    roc_auc_weighted = RocAuc(average='weighted')
    neg_log_loss = EntropyLoss()
    
    
    def __call__(self, y_true: np.ndarray, y_pred_probs: np.ndarray):
        return self.value(y_true, y_pred_probs)
    
    
class SKMultiClfMetricsCalculator:
    def __init__(self, y_true: np.ndarray, y_pred_probs: np.ndarray):
        """_summary_

        :param np.ndarray y_true: ground truth, 1-d array, size (n,)
        :param np.ndarray y_pred_probs: 2-d array, size (n, n_class), the predicted probabilities
            note that the order of the columns should be exactly same as labels in ground truth array
        """
        self.y_true = y_true
        if isinstance(y_pred_probs, pd.DataFrame):
            self.y_pred_probs = y_pred_probs.values
        else:
            self.y_pred_probs = y_pred_probs
        self.num_cls = y_pred_probs.shape[1]
        
    def score(self, metric: SKClfMetricsEnum) -> float:
        """get classification score metric

        :param SKClfMetricsEnum metric: metric defined by SKClfMetricsEnum
        :return float: metric calculated on the given dataset
        """
        return metric(self.y_true, self.y_pred_probs)
    
    def binary_metrics_by_threshold(self, cls_idx: int) -> pd.DataFrame:
        """get binary classification metrics by threshold (1 vs rest approach)

        :param int cls_idx: the class index
        :return pd.DataFrame: DataFrame of columns 
                threshold, tp, fp, tn, fn, tpr, fpr, precision, recall]
        """
        if 0 <= cls_idx <= self.num_cls - 1:
            y_true_binarized = label_binarize(
                self.y_true, 
                classes = list(range(self.y_pred_probs.shape[1]))
            )
            y_true = y_true_binarized[:, cls_idx]
            pred_prob = self.y_pred_probs[:, cls_idx]
            return binary_clf_metric_at_thresholds(y_true, pred_prob)
        else:
            raise ValueError(f"Total {self.num_cls} classes")
        
    def roc_auc_curves(self) -> Dict[int, pd.DataFrame]:
        """roc auc curves

        :return Dict[int, pd.DataFrame]: {
                cls_idx: DataFrame[threshold, tpr, fpr]
            }
        """
        curves = dict()
        for cls_idx in range(self.num_cls):
            metrics = self.binary_metrics_by_threshold(cls_idx)
            curves[cls_idx] = metrics[['threshold', 'tpr', 'fpr']]
        return curves
    
    def roc_auc_curve_agg(self, agg: Literal['micro', 'macro']) -> pd.DataFrame:
        """get averaged roc auc (tpr-fpr)
        
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html#one-vs-rest-multiclass-roc
        
        In a multi-class classification setup with highly imbalanced classes, 
            micro-averaging is preferable over macro-averaging

        :param Literal[&#39;micro&#39;, &#39;macro&#39;] agg: aggregation method
        :return pd.DataFrame: containing [threshold, tpr, fpr]
        """
        y_true_binarized = label_binarize(
            self.y_true, 
            classes = list(range(self.y_pred_probs.shape[1]))
        )
        if agg == 'micro':
            fpr, tpr, threshold = roc_curve(
                y_true_binarized.ravel(), 
                self.y_pred_probs.ravel(),
                drop_intermediate=False
            )

        elif agg == 'macro':
            fprs, tprs, thresholds = dict(), dict(), dict()
            for i in range(self.num_cls):
                fprs[i], tprs[i], thresholds[i] = roc_curve(
                    y_true_binarized[:, i], 
                    self.y_pred_probs[:, i],
                    drop_intermediate=False
                )
    
            threshold = np.linspace(0.0, 1.0, 2500)
            tpr = np.zeros_like(threshold)
            fpr = np.zeros_like(threshold)
            for i in range(self.num_cls):
                # linear interpolation
                # need to be reversed order because second param must be increasing
                tpr += np.interp(threshold, thresholds[i][::-1], tprs[i][::-1]) 
                fpr += np.interp(threshold, thresholds[i][::-1], fprs[i][::-1])
                
            # Average it and compute AUC
            tpr /= self.num_cls
            fpr /= self.num_cls
            
        else:
            raise ValueError("only support micro or macro aggregation")
        
        return pd.DataFrame({
                'threshold': threshold,
                'fpr': fpr,
                'tpr': tpr
            })
        
    def precision_recall_curves(self) -> Dict[int, pd.DataFrame]:
        """precision recall curves

        :return Dict[int, pd.DataFrame]: {
                cls_idx: DataFrame[threshold, precision, recall]
            }
        """
        curves = dict()
        for cls_idx in range(self.num_cls):
            metrics = self.binary_metrics_by_threshold(cls_idx)
            curves[cls_idx] = metrics[['threshold', 'precision', 'recall']]
        return curves
        
    def precision_recall_curve_agg(self, agg: Literal['micro', 'macro']) -> pd.DataFrame:
        """get averaged precision-recall curve
        
        https://scikit-learn.org/stable/auto_examples/model_selection/plot_roc.html#one-vs-rest-multiclass-roc
        
        In a multi-class classification setup with highly imbalanced classes, 
            micro-averaging is preferable over macro-averaging

        :param Literal[&#39;micro&#39;, &#39;macro&#39;] agg: aggregation method
        :return pd.DataFrame: containing [threshold, precision, recall]
        """
        y_true_binarized = label_binarize(
            self.y_true, 
            classes = list(range(self.y_pred_probs.shape[1]))
        )
        if agg == 'micro':
            precision, recall, threshold = precision_recall_curve(
                y_true_binarized.ravel(), 
                self.y_pred_probs.ravel(),
            )
            threshold = np.array(list(threshold) + [1])

        elif agg == 'macro':
            precisions, recalls, thresholds = dict(), dict(), dict()
            for i in range(self.num_cls):
                precisions[i], recalls[i], thresholds[i] = precision_recall_curve(
                    y_true_binarized[:, i], 
                    self.y_pred_probs[:, i],
                )
                # precision and recall have n+1 elements while threshold only have n
                # precision[n+1] = 1, recall[n+1] = 0
                thresholds[i] = np.array(list(thresholds[i]) + [1])
                    
    
            threshold = np.linspace(0.0, 1.0, 2500)
            precision = np.zeros_like(threshold)
            recall = np.zeros_like(threshold)
            for i in range(self.num_cls):
                # linear interpolation
                precision += np.interp(threshold, thresholds[i], precisions[i]) 
                recall += np.interp(threshold, thresholds[i], recalls[i])
                
            # Average it and compute AUC
            precision /= self.num_cls
            recall /= self.num_cls
            
        else:
            raise ValueError("only support micro or macro aggregation")
        
        return pd.DataFrame({
                'threshold': threshold,
                'precision': precision,
                'recall': recall
            })



import ibis
import mlflow

def _ibis_dtype_2_mlflow_dtype_mapping(ibis_dtype: ibis.expr.datatypes.DataType
        ) -> mlflow.types.schema.DataType | mlflow.types.schema.Array | mlflow.types.schema.Map:
    """convert ibis data dtype to mlflow data dtype

    :param ibis.expr.datatypes.DataType ibis_dtype: ibis dtype, see 
        https://ibis-project.org/reference/datatypes#parameters
    :return mlflow.types.schema.DataType | mlflow.types.schema.Array | mlflow.types.schema.Map:
        mlflow support either single value dtype, array or map type, see
        https://mlflow.org/docs/latest/python_api/mlflow.types.html#mlflow.types.DataType
    """
    if ibis_dtype.is_binary():
        return mlflow.types.schema.DataType.binary
    elif ibis_dtype.is_boolean():
        return mlflow.types.schema.DataType.boolean
    elif ibis_dtype.is_date() or ibis_dtype.is_timestamp():
        return mlflow.types.schema.DataType.datetime
    elif ibis_dtype.is_float64() or ibis_dtype.is_decimal():
        return mlflow.types.schema.DataType.double
    elif ibis_dtype.is_float32() or ibis_dtype.is_float16():
        return mlflow.types.schema.DataType.float
    elif ibis_dtype.is_integer(): 
        if ibis_dtype.is_int8() or ibis_dtype.is_int16() or ibis_dtype.is_int32() \
            or ibis_dtype.is_uint8() or ibis_dtype.is_uint16() or ibis_dtype.is_uint32():
            return mlflow.types.schema.DataType.integer
        else:
            return mlflow.types.schema.DataType.long
    elif ibis_dtype.is_array():
        return mlflow.types.schema.Array(
            dtype = _ibis_dtype_2_mlflow_dtype_mapping(ibis_dtype.value_type)
        )
    elif ibis_dtype.is_map():
        if not ibis_dtype.key_type.is_string():
            raise TypeError("Only support string key dtype for Map mlflow schema type")
        return mlflow.types.schema.Map(
            value_type = _ibis_dtype_2_mlflow_dtype_mapping(ibis_dtype.value_type)
        )
    else:
        return mlflow.types.schema.DataType.string


def ibis_dtype_2_mlflow_dtype(colname: str, ibis_dtype: ibis.expr.datatypes.DataType
        ) -> mlflow.types.schema.ColSpec:
    """convert from ibis dtype to mlflow ColSpec (similar to column in database)

    :param str colname: column name
    :param ibis.expr.datatypes.DataType ibis_dtype: ibis data dtype
    :return mlflow.types.schema.ColSpec: mlflow ColSpec
    """
    mlflow_dtype = _ibis_dtype_2_mlflow_dtype_mapping(ibis_dtype)
    
    return mlflow.types.schema.ColSpec(
        name = colname,
        required = not ibis_dtype.nullable,
        type = mlflow_dtype
    )
    
def ibis_schema_2_mlflow_schema(ibis_schema: ibis.expr.schema.Schema
        ) -> mlflow.types.schema.Schema:
    """convert ibis table schema to mlflow table schema

    :param ibis.expr.schema.Schema ibis_schema: ibis table schema
    :return mlflow.types.schema.Schema: mlflow table schema
    """
    mlflow_dtypes = []
    for col, dtype in ibis_schema.items():
        mlflow_dtypes.append(ibis_dtype_2_mlflow_dtype(
            colname = col, 
            ibis_dtype = dtype
        ))
    return mlflow.types.schema.Schema(inputs = mlflow_dtypes)


import ibis
from ibis import _
from typing import Tuple

class SimpleSplitterIbis:
    def __init__(self, shuffle_key: str, train_size: float = 0.8, 
                 random_seed: int = 0, verbose: bool = True):
        """split into train and test set

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param float train_size: train sample size as of total size, defaults to 0.8
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        self.random_seed = random_seed
        self.verbose = verbose
        self.shuffle_key = shuffle_key
        if train_size > 1 or train_size < 0:
            raise ValueError("train_size can only between (0,1)")
        self.train_size = train_size
        
    def split(self, dataset: ibis.expr.types.Table) -> Tuple[ibis.expr.types.Table, ibis.expr.types.Table]:
        """split into train and test set

        :param ibis.expr.types.Table dataset: the ibis dataset
        :return Tuple[ibis.expr.types.Table, ibis.expr.types.Table]: [train_ds, test_ds]
        """
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        dataset = (
            dataset
            .order_by(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
            )
            .mutate(
                (ibis.row_number().over() 
                / _.count().over() > self.train_size)
                .ifelse(0, 1)
                .name('IS_TRAIN_SET_')
            )
        )
        # step 3: split into train/test set
        train_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 1)
            .drop('IS_TRAIN_SET_')
        )
        test_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 0)
            .drop('IS_TRAIN_SET_')
        )
        
        return train_ds, test_ds
        

class StratifiedSplitterIbis(SimpleSplitterIbis):
    def __init__(self, shuffle_key: str, stratify_key: str, train_size: float = 0.8, 
                 random_seed: int = 0, verbose: bool = True):
        """split into train and test set, stratified on given stratify_key

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param str stratify_key: the column name for determine stratify
        :param float train_size: train sample size as of total size, defaults to 0.8
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        super().__init__(
            train_size = train_size,
            shuffle_key = shuffle_key,
            random_seed = random_seed,
            verbose = verbose
        )
        self.stratify_key = stratify_key
        
    def split(self, dataset: ibis.expr.types.Table) -> Tuple[ibis.expr.types.Table, ibis.expr.types.Table]:
        """split into train and test set

        :param ibis.expr.types.Table dataset: the ibis dataset
        :return Tuple[ibis.expr.types.Table, ibis.expr.types.Table]: [train_ds, test_ds]
        """
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        if self.stratify_key not in dataset.columns:
            raise KeyError(f"Stratify key {self.stratify_key} not found in dataset")
        dataset = (
            dataset
            .mutate(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
                .name('RAND_')
            )
            .mutate(
                (
                    (
                    ibis
                    .row_number()
                    .over(
                        ibis.window(
                            group_by = [self.stratify_key],
                            order_by = ['RAND_']
                        )
                    ) # partition by target column -- stratify
                    ) / (
                        _
                        .count()
                        .over(
                            ibis.window(
                                group_by = [self.stratify_key]
                            )
                        )
                    ) > self.train_size
                ).ifelse(0, 1)
                .name('IS_TRAIN_SET_')
            )
        )
        # step 3: split into train/test set
        train_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 1)
            .drop('IS_TRAIN_SET_', 'RAND_')
        )
        test_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 0)
            .drop('IS_TRAIN_SET_', 'RAND_')
        )
        
        return train_ds, test_ds

class GroupSplitterIbis(SimpleSplitterIbis):
    def __init__(self, shuffle_key: str, group_key: str, train_size: float = 0.8,
                 random_seed: int = 0, verbose: bool = True):
        """split into train and test set, each group_key will only appear in either train or test set, not both

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param str group_key: the column name for determine group of each sample
        :param float train_size: train sample size as of total size, defaults to 0.8
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        super().__init__(
            train_size = train_size,
            shuffle_key = shuffle_key,
            random_seed = random_seed,
            verbose = verbose
        )
        self.group_key = group_key
        
    def split(self, dataset: ibis.expr.types.Table) -> Tuple[ibis.expr.types.Table, ibis.expr.types.Table]:
        """split into train and test set

        :param ibis.expr.types.Table dataset: the ibis dataset
        :return Tuple[ibis.expr.types.Table, ibis.expr.types.Table]: [train_ds, test_ds]
        """
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        if self.group_key not in dataset.columns:
            raise KeyError(f"Group key {self.group_key} not found in dataset")
        # assume group_key = CUST_ID, train_size = 0.8
        # step 1: split into train/test group, ensure same group col only exist in either train or test
        groups = (
            # step 1.1 group by CUST_ID so that each CUST_ID will only appear in 1 set
            # select CUST_ID, count() as NUM_ from dataset group by CUST_ID
            dataset
            .group_by(self.group_key)
            .aggregate(
                _.count().name('NUM_')
            )
            # step 1.2 add a random column for ordering and randomness
            # select cast(hash(CUST_ID) as string) as RAND_ from _
            .mutate(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
                .name('RAND_')
            )
            # step 1.3 add a divider between train and test set
            # select CUST_ID, if(sum(NUM_) over (order by RAND_) / sum() over() > 0.25, 0, 1) as IS_TRAIN_SET_
            .mutate(
                (
                    (
                    _['NUM_']
                    .sum()
                    .over(
                        ibis.window(
                            order_by = 'RAND_',
                            following = 0
                        )
                    ) # partition by target column -- stratify
                    ) / (
                        _['NUM_']
                        .sum()
                        .over()
                    ) > self.train_size
                ).ifelse(0, 1)
                .name('IS_TRAIN_SET_')
            )
        )
        groups = (
            dataset
            .left_join(
                groups.select(self.group_key, 'IS_TRAIN_SET_'),
                [self.group_key]
            )
            .drop(f'{self.group_key}_right')
        )
        # step 3: split into train/test set
        train_ds = (
            groups
            .filter(groups['IS_TRAIN_SET_'] == 1)
            .drop('IS_TRAIN_SET_')
        )
        test_ds = (
            groups
            .filter(groups['IS_TRAIN_SET_'] == 0)
            .drop('IS_TRAIN_SET_')
        )
        # step 4: validate that same CUST_ID only in either train or test but not both
        assert (
            train_ds[train_ds[self.group_key].isin(test_ds[self.group_key])]
            .count()
        ).to_pandas() == 0, f"Found some {self.group_key} in both train and test set"
        
        return train_ds, test_ds
    
    
class TimeSeriesSplitterIbis(SimpleSplitterIbis):
    def __init__(self, shuffle_key: str, ts_key: str, train_size: float = 0.8, 
                 random_seed: int = 0, verbose: bool = True):
        """split into train and test set, but test will be later than train on ts_key

        :param str shuffle_key: id column for identitying each row, if given, will shuffle
        :param str ts_key: time series column indicating the order of dataframe
        :param float train_size: train sample size as of total size, defaults to 0.8
        :param int random_seed: control randomness
        :param bool verbose: whether to print out validation message and statistics
        """
        super().__init__(
            train_size = train_size,
            shuffle_key = shuffle_key,
            random_seed = random_seed,
            verbose = verbose
        )
        self.ts_key = ts_key
        
    def split(self, dataset: ibis.expr.types.Table) -> Tuple[ibis.expr.types.Table, ibis.expr.types.Table]:
        """split into train and test set

        :param ibis.expr.types.Table dataset: the ibis dataset
        :return Tuple[ibis.expr.types.Table, ibis.expr.types.Table]: [train_ds, test_ds]
        """
        if self.ts_key not in dataset.columns:
            raise KeyError(f"Time Series key {self.ts_key} not found in dataset")
        if self.shuffle_key not in dataset.columns:
            raise KeyError(f"Shuffle key {self.shuffle_key} not found in dataset")
        dataset = (
            dataset
            .mutate(
                (_[self.shuffle_key]
                .try_cast("string") + str(self.random_seed)) # randomness here
                .hash()
                .name('RAND_')
            )
            .order_by([self.ts_key, 'RAND_'])
            .mutate(
                (ibis.row_number().over() 
                / _.count().over() > self.train_size)
                .ifelse(0, 1)
                .name('IS_TRAIN_SET_')
            )
        )
        train_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 1)
            .drop('IS_TRAIN_SET_', 'RAND_')
        )
        test_ds = (
            dataset
            .filter(dataset['IS_TRAIN_SET_'] == 0)
            .drop('IS_TRAIN_SET_', 'RAND_')
        )
        assert (
            train_ds[self.ts_key].max().to_pandas() <= test_ds[self.ts_key].min().to_pandas()
        ), f"There are some records in train set that have {self.ts_key} later than that in test set"
        
        return train_ds, test_ds


from pathlib import Path
from fsspec import AbstractFileSystem
import yaml
from luntaiDs.CommonTools.dtyper import DSchema


class BaseSchemaManager:
    """base interface for managing table schema metadata
    
    each schema have multiple tables, and each table have a dictionary of schema metadata
    """
    def write(self, schema: str, table: str, dschema: DSchema):
        """save the given table schema into the given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :param DSchema dschema: the table schema object
        """
        content = dschema.to_js()
        self.write_raw(schema, table, content)
    
    def write_raw(self, schema: str, table: str, content: dict):
        """handle how to write raw (dictionary) config into given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :param dict content: the dict version of Dschema object
        """
        raise NotImplementedError("")
    
    def read(self, schema: str, table: str) -> DSchema:
        """read from the system and return the assembled DSchema object

        :param str schema: the schema name
        :param str table: the table name
        :return DSchema: the table schema object
        """
        content = self.read_raw(schema, table)
        return DSchema.from_js(content)
    
    def read_raw(self, schema: str, table: str) -> dict:
        """handle how to write read (dictionary) config from given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :return dict: the dict version of Dschema object 
        """
        raise NotImplementedError("")


class SchemaManagerFileSystem(BaseSchemaManager):
    """Table Schema Manager on fsspec compatible filesystem
    """
    def __init__(self, fs: AbstractFileSystem, root_dir: str):
        """schema manager for file system
        
        folder structure:
        root_dir
            - schema1
                - table A.yml
                - table B.yml
            - schema2
                - table C.yml
                - table D.yml

        :param AbstractFileSystem fs: the fsspec compatible filesystem
        :param str root_dir: root path, if on object storage, 
            the full path including buckets
        """
        self._fs = fs
        self._root_dir = root_dir
        
    def write_raw(self, schema: str, table: str, content: dict):
        """handle how to write raw (dictionary) config into given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :param dict content: the dict version of Dschema object
        """
        # create folder
        schema_folder = Path(self._root_dir) / schema
        self._fs.makedirs(
            path = schema_folder.as_posix(),
            exist_ok = True
        )
        # save content to file
        schema_file = schema_folder / f"{table}.yml"
        with self._fs.open(schema_file, 'w') as obj:
            yaml.dump(
                content, 
                obj, 
                default_flow_style = True, 
                sort_keys = True
            )
            
    def read_raw(self, schema: str, table: str) -> dict:
        """handle how to write read (dictionary) config from given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :return dict: the dict version of Dschema object 
        """
        schema_file = Path(self._root_dir) / schema / f"{table}.yml"
        try:
            with self._fs.open(schema_file, 'r') as obj:
                record = yaml.load(obj, Loader=yaml.Loader)
        except FileNotFoundError as e:
            raise ValueError("No record found for given schema and table")
        else:
            return record
    


from configparser import ConfigParser
import toml
from getpass import getpass
from typing import List
from luntaiDs.CommonTools.settings import SETTINGS

def get_credential(section: str, attr_keys: List[str] = None):
    """get credentials from credential file specified in Settings

    :param section: which section in the .ini file will extract the credentials
    :param attr_keys: a list of keys, by default will be [username, password]
    :return:
    """
    if not attr_keys:
        attr_keys = ['username', 'password']

    if SETTINGS.DB_CREDENTIAL_INTERACTIVE:
        values = []
        for attr in attr_keys:
            v = getpass(prompt = attr)
            values.append(v)
        return values
    else:
        credential_path = SETTINGS.DB_CREDENTIAL_PATH
        if credential_path.endswith(".ini"):
            config = ConfigParser()
            config.read()
            sec = config[section]
        elif credential_path.endswith(".toml"):
            with open(credential_path) as obj:
                config = toml.load(obj)
                sec = config[section]
        else:
            raise TypeError("Only support file format .ini or .toml")
        
        values = []
        for attr in attr_keys:
            v = sec[attr]
            values.append(v)
        return values
    
if __name__ == '__main__':
    SETTINGS.SET_CREDENTIAL_INTERACTIVE()



from typing import Optional, Union, Any, Dict, List, Literal, MutableSequence
from dataclasses import asdict, dataclass
import pandas as pd
import numpy as np
from functools import partial
from sklearn.base import BaseEstimator
from sklearn.feature_selection import mutual_info_classif, SelectKBest
from sklearn.pipeline import Pipeline
from sklearn.preprocessing import OneHotEncoder, RobustScaler, StandardScaler, MaxAbsScaler, \
    PowerTransformer, PowerTransformer, FunctionTransformer
from sklearn.compose import ColumnTransformer
from category_encoders.ordinal import OrdinalEncoder
from category_encoders.count import CountEncoder
from category_encoders.woe import WOEEncoder # bug
from luntaiDs.ModelingTools.Explore.summary import BaseStatObj, BinaryStatObj, NominalCategStatObj, \
    NumericStatObj, OrdinalCategStatAttr, OrdinalCategStatObj, OrdinalCategStatSummary, TabularStat, \
    exp10pc, log10pc
from luntaiDs.ModelingTools.FeatureEngineer.transformers import BucketCategValue, BucketCategByFsel, NamedTransformer, \
    MyImputer, BinaryConverter, OutlierClipper
from luntaiDs.ModelingTools.utils.support import make_present_col_selector


def intStringfy(X: pd.DataFrame) -> pd.DataFrame:
    return (
        X
        .astype('float')
        .astype('Int32')
        .astype('str')
        .replace(pd.NA, np.nan)
        .replace('<NA>', np.nan)
    )

def Stringfy(X: pd.DataFrame) -> pd.DataFrame:
    return X.astype('str')

def filterZero(X: pd.DataFrame) -> pd.DataFrame:
    return X.replace(0, np.nan)

def floatFy(X: pd.DataFrame) -> pd.DataFrame:
    return X.astype('float')

@dataclass
class BaseFeaturePreproc:
    def serialize(self) -> dict:
        return asdict(self)
    
    @classmethod
    def deserialize(cls, attr: dict):
        return cls(**attr)
    
    def compile_sklearn_pipeline(self, param: BaseStatObj) -> Pipeline:
        raise NotImplementedError("")


@dataclass
class BinaryFeaturePreproc(BaseFeaturePreproc):
    def compile_sklearn_pipeline(self, param: BinaryStatObj) -> Pipeline:
        transformers = []

        if param.attr.int_dtype_:
            transformers.append(
                ('int_stringfy', FunctionTransformer(func=intStringfy))
            ) # convert to string type
        
        binarize = BinaryConverter(
            pos_values = list(str(v) for v in param.attr.pos_values_),
            keep_na = True
        )
        imputer = MyImputer(
            strategy = 'constant', 
            fill_value = int(param.attr.na_to_pos_), 
            add_indicator = False
        )

        return Pipeline(
            transformers + [
            ('binarize', binarize),
            ('impute', imputer)
        ])


@dataclass
class OrdinalCategFeaturePreproc(BaseFeaturePreproc):
    impute: bool = True
    standardize: bool = False

    def compile_sklearn_pipeline(self, param: OrdinalCategStatObj) -> Pipeline:
        transformers = []

        if param.attr.int_dtype_:
            transformers.append(
                ('int_stringfy', FunctionTransformer(func=intStringfy))
            ) # convert to string type

        oe =  NamedTransformer(
            OrdinalEncoder(
                cols = [param.colname],
                mapping = [{
                    'col' : param.colname,
                    'mapping' : dict(zip(
                        map(str, param.attr.categories_), 
                        range(len(param.attr.categories_))
                    ))
                }],
                handle_unknown = 'value',
                handle_missing = 'value' if self.impute else 'return_nan'
            )
        )
        transformers.append(('ordinal', oe))
        
        if self.standardize:
            scaler = MaxAbsScaler()
            transformers.append(
                ('scale', NamedTransformer(scaler))
            )
        return Pipeline(transformers)

@dataclass
class NominalCategFeaturePreproc(BaseFeaturePreproc):
    impute_value: Optional[str] = 'Other'
    bucket_strategy: Literal['freq', 'correlation'] = None
    encode_strategy: Optional[Literal['ohe', 'ce', 'woe']] = 'woe'
    
    def compile_sklearn_pipeline(self, param: NominalCategStatObj) -> Pipeline:
        transformers = []

        if self.impute_value is not None:
            if self.impute_value == 'most_frequent':
                imputer = MyImputer(
                    strategy = 'most_frequent',
                    add_indicator = False
                )
            else:
                imputer = MyImputer(
                    strategy = 'constant', 
                    fill_value = self.impute_value,
                    add_indicator = False
                )
            transformers.append(('impute', imputer))

        if param.attr.int_dtype_:
            transformers.append(
                ('int_stringfy', FunctionTransformer(func=intStringfy))
            ) # convert to string type
        else:
            transformers.append(
                ('stringfy', FunctionTransformer(func=Stringfy))
            ) # convert to string type

        if param.summary.unique_.value > param.attr.max_categories_:
            if self.bucket_strategy == 'freq':
                # freq threshold
                freq_threshold = param.summary.vpercs_.cumsum().iloc[param.attr.max_categories_ - 1]
                bucketizer = BucketCategValue(
                    threshold = freq_threshold,
                    handle_unknown = 'other' # will use `Other` indeed
                )
            else:
                bucketizer = BucketCategByFsel(
                    fsel = SelectKBest(
                        score_func = partial(
                            mutual_info_classif, 
                            discrete_features = True
                        ),
                        k = param.attr.max_categories_
                    ),
                    other_lvl_name = 'Other'
                )
            transformers.append(('bucketize', bucketizer))

        if self.encode_strategy is not None:
            if self.encode_strategy == 'ohe':
                encoder = OneHotEncoder(
                    sparse=False, 
                    handle_unknown='ignore'
                )
            elif self.encode_strategy == 'ce':
                encoder = CountEncoder(
                    handle_unknown='value', 
                    handle_missing='value',
                    normalize = True,
                    combine_min_nan_groups=True,
                )
            elif self.encode_strategy == 'woe':
                encoder = WOEEncoder(
                    handle_unknown='value', 
                    handle_missing='value',
                    randomized=True, # prevent overfitting
                    sigma=0.05
                )

            transformers.append(
                ('encode', NamedTransformer(encoder))
            )

        return Pipeline(transformers)


@dataclass
class NumericFeaturePreproc(BaseFeaturePreproc):
    impute: bool = True
    normalize:bool = False
    standardize_strategy: Optional[Literal['robust', 'standard', 'maxabs']] = 'robust'
    
    def compile_sklearn_pipeline(self, param: NumericStatObj) -> Pipeline:
        transformers = []

        transformers.append(
            ('float', FunctionTransformer(func=floatFy))
        ) # convert to float type

        if param.attr.xtreme_method_ is not None:
            if param.attr.setaside_zero_:
                clipper = OutlierClipper(
                    strategy = param.attr.xtreme_method_,
                    quantile_range = (1, 99),
                    filter_func = filterZero
                )
            else:
                clipper = OutlierClipper(
                    strategy = param.attr.xtreme_method_,
                    quantile_range = (1, 99),
                )
            transformers.append(('clip', clipper))

        if self.impute:
            if param.attr.setaside_zero_:
                imputer = MyImputer(
                    strategy = 'constant',
                    fill_value = 0,
                    add_indicator = False
                )
            else:
                if param.summary.stat_descriptive_.skew is None:
                    strategy = 'median'
                elif abs(param.summary.stat_descriptive_.skew) > 0.5:
                    strategy = 'median'
                else:
                    strategy = 'mean'
                imputer = MyImputer(
                    strategy = strategy,
                    add_indicator = False
                )
            transformers.append(('impute', imputer))

        if param.attr.log_scale_:
            logger = NamedTransformer(
                FunctionTransformer(
                    func = log10pc, 
                    inverse_func = exp10pc
                )
            )
            transformers.append(('log', logger))

        if self.normalize:
            normalizer = NamedTransformer(
                PowerTransformer(
                    method = 'yeo-johnson',
                    standardize = False
                )
            )
            transformers.append(('normalize', normalizer))

        if self.standardize_strategy is not None:
            if self.standardize_strategy == 'robust':
                scaler = RobustScaler()
            elif self.standardize_strategy == 'maxabs':
                # will be between -1 and 1
                scaler = MaxAbsScaler()
            else:
                scaler = StandardScaler()
            transformers.append(('scale', NamedTransformer(scaler)))

        
        return Pipeline(transformers)


def get_preliminary_preprocess(ts: TabularStat) -> BaseEstimator:
    transformers = []    
    for col, stat in ts.items():
        if col in ts.get_nominal_cols():
            transformer = NominalCategFeaturePreproc(
                impute_value = 'Other',
                bucket_strategy = None,
                encode_strategy = 'woe'
            ).compile_sklearn_pipeline(stat)
        elif col in ts.get_ordinal_cols():
            transformer = OrdinalCategFeaturePreproc(
                impute = True,
                standardize = True
            ).compile_sklearn_pipeline(stat)
        elif col in ts.get_binary_cols():
            transformer = BinaryFeaturePreproc(
            ).compile_sklearn_pipeline(stat)
        elif col in ts.get_numeric_cols():
            transformer = NumericFeaturePreproc(
                impute = True,
                normalize = False,
                standardize_strategy = 'robust'
            ).compile_sklearn_pipeline(stat)
        
        transformers.append(
            (col, transformer, make_present_col_selector([col]))
        )
        
    pipe = NamedTransformer(
        ColumnTransformer(
            transformers,
            remainder='drop'
        )
    )
    return pipe


def get_mutual_info_preprocess(ts: TabularStat) -> BaseEstimator:
    transformers = []    
    for col, stat in ts.items():
        if col in ts.get_nominal_cols():
            # switch to Norminal stat
            stat_ord = OrdinalCategStatObj(
                colname = col,
                attr = OrdinalCategStatAttr(
                    categories_ = stat.summary.vcounts_.index.tolist(),
                    int_dtype_ = stat.attr.int_dtype_
                ),
                summary = OrdinalCategStatSummary(
                    colname_ = col,
                    total_ = stat.summary.total_,
                    missing_ = stat.summary.missing_,
                    unique_ = stat.summary.unique_,
                    vcounts_ = stat.summary.vcounts_,
                    vpercs_ = stat.summary.vpercs_,
                )
            )
            transformer = OrdinalCategFeaturePreproc(
                impute = True,
                standardize = False
            ).compile_sklearn_pipeline(stat_ord)
        elif col in ts.get_ordinal_cols():
            transformer = OrdinalCategFeaturePreproc(
                impute = True,
                standardize = False
            ).compile_sklearn_pipeline(stat)
        elif col in ts.get_binary_cols():
            transformer = BinaryFeaturePreproc(
            ).compile_sklearn_pipeline(stat)
        elif col in ts.get_numeric_cols():
            transformer = NumericFeaturePreproc(
                impute = True,
                normalize = False,
                standardize_strategy = 'robust'
            ).compile_sklearn_pipeline(stat)
        
        transformers.append(
            (col, transformer, make_present_col_selector([col]))
        )
        
    pipe = NamedTransformer(
        ColumnTransformer(
            transformers,
            remainder='drop'
        )
    )
    return pipe

"""preprocessing model"""

@dataclass
class BaseFeaturePreprocModel:
    stat_obj: BaseStatObj
    preproc: BaseFeaturePreproc
    
    def compile_sklearn_pipeline(self) -> Pipeline:
        return self.preproc.compile_sklearn_pipeline(
            param = self.stat_obj
        )
        
    def serialize(self) -> dict:
        return dict(
            constructor = self.__class__.__name__,
            colname = self.stat_obj.colname,
            stat_obj = self.stat_obj.serialize(),
            preproc = self.preproc.serialize()
        )
    
    @classmethod
    def deserialize(cls, model_dict: dict):
        raise NotImplementedError("")
        
@dataclass
class BinaryFeaturePreprocModel(BaseFeaturePreprocModel):
    stat_obj: BinaryStatObj
    preproc: BinaryFeaturePreproc = BinaryFeaturePreproc()
    
    @classmethod
    def deserialize(cls, model_dict: dict):
        return cls(
            stat_obj = BinaryStatObj.deserialize(model_dict['stat_obj']),
            preproc = BinaryFeaturePreproc.deserialize(model_dict['preproc'])
        )
        
@dataclass
class OrdinalFeaturePreprocModel(BaseFeaturePreprocModel):
    stat_obj: OrdinalCategStatObj
    preproc: OrdinalCategFeaturePreproc = OrdinalCategFeaturePreproc()
    
    @classmethod
    def deserialize(cls, model_dict: dict):
        return cls(
            stat_obj = OrdinalCategStatObj.deserialize(model_dict['stat_obj']),
            preproc = OrdinalCategFeaturePreproc.deserialize(model_dict['preproc'])
        )
        
@dataclass
class NominalFeaturePreprocModel(BaseFeaturePreprocModel):
    stat_obj: NominalCategStatObj
    preproc: NominalCategFeaturePreproc = NominalCategFeaturePreproc()
    
    @classmethod
    def deserialize(cls, model_dict: dict):
        return cls(
            stat_obj = NominalCategStatObj.deserialize(model_dict['stat_obj']),
            preproc = NominalCategFeaturePreproc.deserialize(model_dict['preproc'])
        )
        
@dataclass
class NumericFeaturePreprocModel(BaseFeaturePreprocModel):
    stat_obj: NumericStatObj
    preproc: NumericFeaturePreproc = NumericFeaturePreproc()
    
    @classmethod
    def deserialize(cls, model_dict: dict):
        return cls(
            stat_obj = NumericStatObj.deserialize(model_dict['stat_obj']),
            preproc = NumericFeaturePreproc.deserialize(model_dict['preproc'])
        )
    

class TabularPreprocModel(
        list, 
        MutableSequence[BaseFeaturePreprocModel]
    ):
    CONSTRUCTORS = {
        NumericFeaturePreprocModel.__name__: NumericFeaturePreprocModel,
        NominalFeaturePreprocModel.__name__: NominalFeaturePreprocModel,
        OrdinalFeaturePreprocModel.__name__: OrdinalFeaturePreprocModel,
        BinaryFeaturePreprocModel.__name__: BinaryFeaturePreprocModel,
    }
    STAT_OBJ_MAPPING = {
        NumericStatObj.__name__: NumericFeaturePreprocModel,
        NominalCategStatObj.__name__: NominalFeaturePreprocModel,
        OrdinalCategStatObj.__name__: OrdinalFeaturePreprocModel,
        BinaryStatObj.__name__: BinaryFeaturePreprocModel,
    }
    """Tabular preprocessing model for multiple columns"""
    
    def serialize(self) -> list:
        return [
            model.serialize() for model in self
        ]
    
    @classmethod
    def deserialize(cls, model_serialized: list):
        r = cls()
        for model in model_serialized:
            c = cls.CONSTRUCTORS.get(model['constructor']) # constructor
            r.append(c.deserialize(model))
        return r
    
    @classmethod
    def from_tabular_stat(cls, ts: TabularStat):
        r = cls()
        for col, obj in ts.items():
            c = cls.STAT_OBJ_MAPPING.get(obj.__class__.__name__) # constructor
            model = c(stat_obj = obj)
            r.append(model)
        return r
    
    def to_tabular_stat(self) -> TabularStat:
        ts = TabularStat()
        for preproc_model in self:
            col = preproc_model.stat_obj.colname
            ts[col] = preproc_model.stat_obj
        return ts
    
    def compile_sklearn_pipeline(self) -> Pipeline:
        transformers = []    
        for preproc_model in self:
            col = preproc_model.stat_obj.colname
            transformer = preproc_model.compile_sklearn_pipeline()
            transformers.append(
                (col, transformer, make_present_col_selector([col]))
            )
            
        pipe = NamedTransformer(
            ColumnTransformer(
                transformers,
                remainder='drop'
            )
        )
        return pipe


import os
from datetime import timedelta
from typing import Dict, List
import numpy as np
import pandas as pd
from itertools import cycle
from bokeh.plotting import figure, show, Figure
from bokeh.models import ColumnDataSource, Dropdown, HoverTool, TableColumn, DataTable, NumeralTickFormatter, Range1d, \
    LinearAxis, Legend, CustomJS, Slider, Column, LinearColorMapper, ColorBar, BasicTicker, Row, Select
from bokeh.models import Panel, Tabs, DateFormatter, Span
from bokeh.models.widgets import CheckboxButtonGroup
from bokeh.layouts import column, row, layout, gridplot
from bokeh.palettes import Spectral, Category20c, Dark2_5, Greys256, Magma256, Dark2_8, Colorblind
from bokeh.transform import factor_cmap, transform

def table_nonedit_general(df: pd.DataFrame, index_col: str = None, dt_columns:List[str] = None, index_position: int=None):
    df = df.copy()
    if index_col is not None:
        df = df.set_index(index_col)
    dt_columns = [] if dt_columns is None else dt_columns
    for col in dt_columns:
        df[col] = pd.to_datetime(df[col])
    source = ColumnDataSource(df)
    columns = [TableColumn(field=col, title = col, formatter = DateFormatter()) for col in dt_columns]
    columns.extend([TableColumn(field=col, title=col) for col in df.columns if col not in dt_columns])
    data_table = DataTable(source=source, columns=columns, editable=False,
                           index_position=index_position, sizing_mode="stretch_both")
    return data_table

def chart_coef_fimpt(df: pd.DataFrame, title: str, size: tuple = (1000, 500)) -> Figure:

    """Coefficient or Feature importance chart

    :param df: 1 row per feature , columns: [Features, Coef], optional column [Color]
    :param title: chart title
    :param size: figsize
    :return:
    """
    colors = cycle(Dark2_5)
    if 'Color' not in df.columns:
        df = df.copy()
        df.loc[:, 'Color'] = [c for _, c in zip(range(len(df)), colors)]

    # df[Features, Coef, Color]
    source_coef = ColumnDataSource(df)
    fig_coef = figure(plot_width = size[0], plot_height = size[1],
                      x_range = source_coef.data['Features'], title=title)
    fig_coef.vbar(
        source = source_coef, x = 'Features', top = 'Coef', bottom = 0,
        width = 0.5, line_color = 'black', fill_color = 'Color'
    )
    fig_coef.add_tools(HoverTool(tooltips=[("Features", "@Features"), ("Coef", "@Coef")]))
    fig_coef.xaxis.major_label_orientation = "vertical"
    return fig_coef

def table_hypt_distr(study_info: dict):
    """Hyperparameter distribution table chart

    :param study_info: dict, {
            'best_params': study.best_params, # dict
            'best_value': study.best_value, # float
            'param_distr': r_distr, # list of dict
            'trial_valus': trial_results, # list
            'hypt_importance': importance_
        }
    :return:
    """
    s = pd.DataFrame.from_records(study_info['param_distr'])
    s['Best'] = s['Hypt'].map(study_info['best_params'])
    s['Importance'] = s['Hypt'].map(study_info['hypt_importance'])
    head_cols = ['Hypt', 'Best', 'Importance', 'Distribution']
    s = s[head_cols + s.columns.difference(head_cols).tolist()].sort_values(by = 'Importance', ascending = False)

    data_table = table_nonedit_general(df = s, index_col='Hypt', index_position=None)
    return data_table


def chart_hypt_imp(hypt_importance: dict, title: str='Hyperparameter Importance'):
    """Hyperparameter importance Chart

    :param hypt_importance: dict, apis.get_hypt_info()['hypt_importance']
        {
            hyper_param1: v1,
            hyper_param2: v2,
        }
    :param title: chart title
    :return:
    """
    df = pd.DataFrame(hypt_importance.items())
    df.columns = ['Hypt', 'Importance']
    source_hypt = ColumnDataSource(df.sort_values("Importance"))
    fig_hypt = figure(plot_width=500, plot_height=500, y_range=source_hypt.data['Hypt'],
                      title=title)
    fig_hypt.hbar(
        source=source_hypt, y='Hypt', right='Importance',
        width=0.4, left=0,
        line_color='black',
        #fill_color='Color'
    )
    fig_hypt.add_tools(HoverTool(tooltips=[("Hypt", "@Hypt"), ("Importance", "@Importance")]))
    return fig_hypt


def chart_optbin(bin_r: pd.DataFrame, metric: str = 'Event rate', size: tuple = (1000, 500), title: str = "Realized Event vs. Bins") -> Figure:

    """Optimal binning result chart

    :param bin_r: results returned by optbin.binning_table.build(add_totals=False)
                    columns [Bin, Count, Count (%), Event, Non-event, Event rate, WoE, IV, JS]
    :param metric: metric to be displayed on right axis, default Event rate, can also be WoE, IV, JS
    :param size: figsize
    :return:
    """

    source_bin = ColumnDataSource(bin_r)
    fig_bin = figure(plot_width=size[0], plot_height=size[1], x_range=source_bin.data['Bin'], title = title)
    fig_bin.add_layout(Legend(), 'right')
    fig_bin.vbar_stack(
        ['Non-event', 'Event'], legend_label=['Non-event', 'Event'],
        source = source_bin, x = 'Bin', width = 0.8,
        line_color = 'black', color = ['#6EA3C6', '#CB4932']
    )
    fig_bin.yaxis.formatter = NumeralTickFormatter(format='0.0a')  # million
    # add event rate to secondary axis
    #fig_bin.y_range = Range1d(start = 0, end = 1)
    fig_bin.extra_y_ranges = {"BAD_RT_AXIS": Range1d(start=0, end = 1.2)}
    fig_bin.add_layout(LinearAxis(y_range_name="BAD_RT_AXIS"), 'right')
    fig_bin.line(x='Bin', y=metric, source=source_bin, y_range_name='BAD_RT_AXIS', width=3,
                    legend_label=metric, color='#062335')
    fig_bin.circle(x='Bin', y=metric, source=source_bin, y_range_name='BAD_RT_AXIS', color='black',
                      fill_color="white", size=10)
    fig_bin.add_tools(HoverTool(tooltips=[("Bin", "@Bin"), ("Count", "@Count"), ("Count (%)", "@{Count (%)}{(0.000)}"),
                                          ("Event", "@Event"), ("Non-event", "@{Non-event}"), ("Event rate", "@{Event rate}{(0.000)}"),
                                          ("WoE", "@WoE"), ("IV", "@IV"), ("JS", "@JS")]))
    fig_bin.xaxis.major_label_orientation = "vertical"

    return fig_bin


def chart_roc_curve(*metric_dfs: pd.DataFrame, names:List[str] = None, 
            size: tuple = (1000, 500), title ='ROC Curve') -> layout:
    """ROC curve on different models for comparison

    :param metric_dfs: metrics at each threshold, columns: [threshold, tpr, fpr]
    :param names: names for each metric dfs, use as labels in the chart to differientiate each curve
    :param size: figsize
    :param title: title of the chart
    :return: ROC curve with a slider for threshold
    """
    colors = cycle(Dark2_5)
    fig_roc = figure(plot_width=size[0], plot_height=size[1], title=title,
                     x_axis_label='FPR', y_axis_label='TPR')
    fig_roc.add_layout(Legend(), 'right')
    for metric_df, name in zip(metric_dfs, names):
        source_roc = ColumnDataSource(metric_df)
        fig_roc.line("fpr", "tpr", source = source_roc, color=next(colors), width=3, legend_label=name)
    fig_roc.add_tools(HoverTool(tooltips=[("threshold", "@threshold"), ("fpr", "@fpr"), ("tpr", "@tpr")]))

    with open(os.path.join(os.path.dirname(__file__), "chart_api_js/roc_standalone_callback.js"), "r") as obj:
        callback_ = obj.read()
        
    slider = Slider(start=0.0, end=1.0, value=1, step=0.0001, 
        title='Threshold', format='0[.]0000')
    for slider_pos, slider_df in enumerate(metric_dfs):
        # will add a slider to the relevant curve
        df_anno = pd.DataFrame({
            'threshold': [1], 
            'tpr': [0], 
            'fpr' : [0]
        })
        source_anno = ColumnDataSource(df_anno)
        fig_roc.circle("fpr", "tpr", source = source_anno, 
            color='black', fill_color="#2B699D", size = 10)
       
        callback = CustomJS(
            args=dict(
                source_roc_test=ColumnDataSource(slider_df), 
                source_anno=source_anno
            ),
            code=callback_
        )
        slider.js_on_change('value', callback)
    
    layout = Column(slider, fig_roc)
    return layout


def chart_pr_curve(*metric_dfs: pd.DataFrame, names:List[str] = None, 
        size: tuple = (1000, 500), title ='PR Curve') -> layout:
    """Precision - Recall curve on different models for comparison

    :param metric_dfs: metrics at each threshold, columns: [threshold, tpr, fpr]
    :param names: names for each metric dfs, use as labels in the chart to differientiate each curve
    :param size: figsize
    :param title: title of the chart
    :return: PR curve with a slider for threshold
    """
    colors = cycle(Dark2_8);next(colors);next(colors)
    fig_pr = figure(plot_width=size[0], plot_height=size[1], title=title,
                     x_axis_label='Recall', y_axis_label='Precision')
    fig_pr.add_layout(Legend(), 'right')
    for metric_df, name in zip(metric_dfs, names):
        source_roc = ColumnDataSource(metric_df)
        fig_pr.line("recall", "precision", source=source_roc, color=next(colors), width=3, legend_label=name)
    fig_pr.add_tools(HoverTool(tooltips=[("threshold", "@threshold"), ("recall", "@recall"), ("precision", "@precision")]))

    with open(os.path.join(os.path.dirname(__file__), "chart_api_js/pr_standalone_callback.js"), "r") as obj:
        callback_ = obj.read()
    
    slider = Slider(start=0.0, end=1.0, value=1, step=0.0001, 
            title='Threshold', format='0[.]0000')
    
    for slider_pos, slider_df in enumerate(metric_dfs):
        # will add a slider to the relevant curve
        df_anno = pd.DataFrame({
            'threshold': [1], 
            'recall': [0], 
            'precision' : [1]
        })
        source_anno = ColumnDataSource(df_anno)
        fig_pr.circle("recall", "precision", source=source_anno, 
            color='black', fill_color="#2B699D", size=10)

        callback = CustomJS(
            args=dict(
                source_pr_test=ColumnDataSource(slider_df), 
                source_anno=source_anno
            ),
            code=callback_
        )
        
        slider.js_on_change('value', callback)
    
    layout = Column(slider, fig_pr)
    return layout


def chart_pr_threshold(metrics: pd.DataFrame, size: tuple = (1000, 500), title: str ='PR Tradeoff Curve') -> Figure:
    """Precision Recall tradeoff curve @ each different threshold

    :param metrics: metrics at each threshold, columns: [threshold, precision, recall]
    :param size: figsize
    :param title: title of the chart
    :return:
    """
    source_pr = ColumnDataSource(metrics)
    fig_pr = figure(plot_width=size[0], plot_height=size[1], title=title,
                    x_axis_label='Threshold', y_axis_label='Precision/Recall')
    fig_pr.add_layout(Legend(), 'right')
    fig_pr.line("threshold", "recall", source=source_pr, color='#2EA496', width=3, legend_label='Recall')
    fig_pr.line("threshold", "precision", source=source_pr, color='#CB6040', width=3, legend_label='Precision')
    fig_pr.add_tools(HoverTool(tooltips=[("threshold", "@threshold"), ("recall", "@recall"), ("precision", "@precision")]))
    return fig_pr


def chart_confusion_maxtrix(metrics: pd.DataFrame, size: tuple = (1000, 500), title: str ='Confusion Matrix') -> layout:
    """Confusion matrix at each threshold (Count + Error chart)

    :param metrics: metrics at each threshold, columns: [threshold, fp, tp, tn, fn]
    :param size: figsize
    :param title: title of the chart
    :return: Confusion matrix (count view + error view) and slider for threshold
    """
    initial = metrics.iloc[len(metrics) // 2, :]
    total = metrics.loc[0, ['fp', 'tp', 'tn', 'fn']].sum()

    df_origin = pd.DataFrame(
        [[initial['fp'], initial['tp']], [initial['tn'], initial['fn']]],
        columns=['neg', 'pos'],
        index=['pos', 'neg']
    )
    df_origin.index.name = 'Prediction'
    df_origin.columns.name = 'Actual'
    # Prepare data.frame in the right format
    df_error = pd.DataFrame(
        [[initial['fpr'], 0], [0, 1 - initial['tpr']]],
        columns=['neg', 'pos'],
        index=['pos', 'neg']
    )
    df_error.index.name = 'Prediction'
    df_error.columns.name = 'Actual'
    df_origin = df_origin.stack().rename("value").reset_index()
    df_error = df_error.stack().rename("value").reset_index()
    # You can use your own palette here

    # colors = ['#d7191c', '#fdae61', '#ffffbf', '#a6d96a', '#1a9641']
    origin_source = ColumnDataSource(df_origin)
    error_source = ColumnDataSource(df_error)
    raw_source = ColumnDataSource(metrics)

    # Had a specific mapper to map color with value
    mapper_origin = LinearColorMapper(palette=Magma256, low=0, high=total)
    mapper_error = LinearColorMapper(palette=Greys256, low=0.0, high=1.0)

    # Define a figure

    p_origin = figure(
        plot_width=size[0] // 2, plot_height=size[1], title=f"{title} - Original",
        x_range=['neg', 'pos'], y_range=['pos', 'neg']
    )

    # Create rectangle for heatmap
    p_origin.rect(
        x="Prediction", y="Actual", source=origin_source,
        width=1, height=1, line_color=None, fill_color=transform('value', mapper_origin)
    )

    # Add legend
    color_bar_origin = ColorBar(
        color_mapper=mapper_origin,
        location=(0, 0),
        ticker=BasicTicker(desired_num_ticks=256)
    )
    p_origin.add_layout(color_bar_origin, 'right')
    p_origin.add_tools(HoverTool(tooltips=[("Prediction", "@Prediction"), ("Actual", "@Actual"), ("value", "@value")]))


    p_error = figure(
        plot_width=size[0] // 2, plot_height=size[1], title=f"{title} - Error",
        x_range=['neg', 'pos'], y_range=['pos', 'neg']
    )

    # Create rectangle for heatmap
    p_error.rect(
        x="Prediction", y="Actual", source=error_source,
        width=1, height=1, line_color=None, fill_color=transform('value', mapper_error)
    )

    # Add legend
    color_bar_error = ColorBar(
        color_mapper=mapper_error,
        location=(0, 0),
        ticker=BasicTicker(desired_num_ticks=256)
    )
    p_error.add_layout(color_bar_error, 'right')
    p_error.add_tools(HoverTool(tooltips=[("Prediction", "@Prediction"), ("Actual", "@Actual"), ("value", "@value")]))

    slider = Slider(start=0.0, end=1.0, value=initial['threshold'], step=0.001, title='Threshold', format = '0[.]000')

    with open(os.path.join(os.path.dirname(__file__), "chart_api_js/confunsion_matrix_callback.js"), "r") as obj:
        callback = obj.read()
    callback = CustomJS(
        args=dict(origin_source=origin_source, error_source=error_source, raw_source=raw_source),
        code=callback
    )
    slider.js_on_change('value', callback)

    layout = Column(slider, Row(p_origin, p_error))
    return layout



def chart_metric_score_compare(scores: dict, size: tuple = (1000, 400), title: str ='Metric by model') -> Figure:
    """metric bar chart of each model/submodel

    :param scores: dictionary of metrics, keys are model names, values are corresponding metric
    :param size: figsize
    :param title: chart title
    :return:

    """
    colors = cycle(Dark2_8);next(colors);next(colors);next(colors)
    df = pd.DataFrame({
        'Model': scores.keys(), 
        'Metric': scores.values(), 
        'Color': [c for _, c in zip(scores.keys(), colors)]
    })

    #df.loc[df['Metric'] < df['Metric'].quantile(0.3), 'Color'] = '#DE5644'

    source_metrics = ColumnDataSource(df)

    fig_metric = figure(plot_width=size[0], plot_height=size[1], title=title,
                        x_range=source_metrics.data['Model'])
    fig_metric.add_layout(Legend(), 'right')
    fig_metric.vbar(
        source=source_metrics, x='Model', top='Metric',
        width=0.5, bottom=0,
        line_color='black', fill_color='Color', legend_field="Model"
    )
    fig_metric.add_tools(HoverTool(tooltips=[("Model", "@Model"), ("Metric", "@Metric")]))
    #fig_metric.xaxis.major_label_orientation = 3.14/4
    fig_metric.xaxis.visible = False
    return fig_metric

def chart_calibration_curve(cali_curves: Dict[str, Dict[str, list]], size: tuple = (1200, 400), title: str ='Calibration Curve') -> Figure:
    """plot calibration curve for multiple models

    :param cali_curves: assemble of results from sklearn.calibration_curve
        format cali_curves = {
            'cali1' : {
                # from calibration_curve
                'prob_true': [],
                'prob_pred': [],
            },
            'cali2': {
                # from calibration_curve
                'prob_true': [],
                'prob_pred': [],
            },
            ...
        }
    :param size:
    :param title:
    :return:

    """
    colors = cycle(Dark2_5)
    fig_cali = figure(plot_width=size[0], plot_height=size[1], title=title,
                      x_axis_label='Mean predicted probability', y_axis_label='Fraction of Positives')
    fig_cali.add_layout(Legend(), 'right')
    fig_cali.line(x=[0, 1], y=[0, 1], color='#7A7C77', width=2, line_dash="4 4", legend_label='Perfect Calibrate')
    for cali_name, cali_curve in cali_curves.items():
        source_cali =  ColumnDataSource(cali_curve)
        fig_cali.line("prob_pred", "prob_true", source=source_cali, color=next(colors), width=3, legend_label=cali_name)
        fig_cali.circle("prob_pred", "prob_true", source=source_cali, color='black', fill_color="white", size=10)

    fig_cali.add_tools(HoverTool(tooltips=[("prob_pred", "@prob_pred"), ("prob_true", "@prob_true")]))
    return fig_cali




import numpy as np
import pandas as pd
from typing import List, Dict, Optional
from dataclasses import asdict
from itertools import cycle
from bokeh.plotting import figure, show, Figure
from bokeh.models import ColumnDataSource, Dropdown, HoverTool, TableColumn, DataTable, NumeralTickFormatter, Range1d, \
    LinearAxis, Legend, CustomJS, Slider, ScientificFormatter, StringFormatter, LinearColorMapper, ColorBar, BasicTicker
from bokeh.models import Panel, Tabs, DateFormatter, Span, LegendItem, Legend, Div, Panel, Tabs
from bokeh.layouts import column, row, layout, gridplot
from bokeh.palettes import Dark2_5, Dark2_8, Magma256
from bokeh.transform import factor_cmap, transform
from bokeh.io import save, output_file
from luntaiDs.ModelingTools.Explore.summary import BaseUniVarClfTargetCorr, BinaryStatObj, CategStatObj, \
    CategUniVarClfTargetCorr, DescStat, NumericStatObj, NumericUniVarClfTargetCorr, \
        QuantileStat, StatVar, TabularStat

def _sort_combine(arr: pd.Series, max_item:int = 10) -> pd.Series:
    """sort and combine small categories

    :param pd.Series arr: the value series where value is something metric like count
    :param int max_item: maximum category, if more will merge, defaults to 10
    :return pd.Series: sorted and merged series
    """
    arr = arr.sort_values(ascending = False)
    if len(arr) > max_item:
        arr.iloc[max_item] = arr.iloc[max_item :].sum()
        arr = arr.iloc[:max_item + 1]
        arr.index = arr.index[:-1].tolist() + ['Other']

    return arr


def chart_donut(donut_arr: pd.Series, max_donut:int = 6, colors: list = None, 
        size: tuple = (500, 400), title: str='Donut chart') -> Figure:
    """generate a bokeh donut chart for proportion counts

    :param pd.Series donut_arr: the count series (index = categ, value = counts)
    :param int max_donut: maximum category, if more will merge, defaults to 6
    :param list colors: colors of each category, defaults to None
    :param tuple size: size of the chart, defaults to (500, 400)
    :param str title: title of the chart, defaults to 'Donut chart'
    :return Figure: bokeh donut chart
    """
    if colors is None:
        donut_arr = _sort_combine(donut_arr, max_item = max_donut)
    perc = donut_arr / donut_arr.sum()
    angles = (
        perc
        .map(lambda x: 2 * 3.1415926 * x) # convert to float
        .cumsum()
    )
    color = cycle(Dark2_5)
    donut_dict = pd.DataFrame({
        'label' : donut_arr.index,
        'value' : donut_arr.values,
        'perc' : perc,
        'start' : [0] + angles[:-1].values.tolist(),
        'end' : angles.values,
    })
    if colors is not None:
        donut_dict['colors'] = colors
    else:
        donut_dict['colors'] = donut_dict['label'].map(
            lambda x: next(color) if x != 'Other' else '#9C9C9C'
        )
    
    source = ColumnDataSource(donut_dict)
    p = figure(plot_width=size[0], plot_height=size[1], title=title)
    p.add_layout(Legend(), 'right')
    p.annular_wedge(
        x=0, y=0, source = source,
        inner_radius=0.5, outer_radius=0.85,
        start_angle="start", end_angle="end", legend_group = 'label',
        line_color="black", line_width=2, fill_color='colors'
    )
    p.x_range = Range1d(start=-1, end = 1)
    p.y_range = Range1d(start=-1, end = 1)
    p.add_tools(
        HoverTool(
            tooltips=[
                ("label", "@label"), 
                ("value", "@value"), 
                ("perc", "@perc{(0.0%)}")
            ]
        )
    )
    p.axis.visible = False
    p.grid.visible = False
    return p

def chart_barchart(bar_arr: pd.Series, max_bar:int = 10, size: tuple = (500, 400), 
        title: str='Bar chart') -> Figure:
    """generate a bokeh bar chart for proportion counts

    :param pd.Series bar_arr: the count series (index = categ, value = counts)
    :param int max_bar: maximum category, if more will merge, defaults to 10
    :param tuple size: size of the chart, defaults to (500, 400)
    :param str title: title of the chart, defaults to 'Donut chart'
    :return Figure: bokeh bar chart
    """
    bar_arr = _sort_combine(bar_arr, max_item = max_bar)
    perc = bar_arr / bar_arr.sum()
    color = cycle(Dark2_8)
    bar_dict = pd.DataFrame({
        'label' : bar_arr.index.astype('string').tolist(),
        'value' : bar_arr.values,
        'perc' : perc
    })
    bar_dict['colors'] = bar_dict['label'].map(
        lambda x: next(color) if x != 'Other' else '#9C9C9C'
    )
    
    source = ColumnDataSource(bar_dict)
    p = figure(plot_width=size[0], plot_height=size[1], title=title,
            y_range = source.data['label'])
    p.add_layout(Legend(), 'right')
    p.hbar(
        y='label', right='value', left = 0, source = source,
        height=0.75, legend_group = 'label',
        line_color="black", line_width=2, fill_color='colors'
    )
    p.add_tools(
        HoverTool(
            tooltips=[
                ("label", "@label"), 
                ("value", "@value"), 
                ("perc", "@perc{(0.0%)}")
            ]
        )
    )
    #p.yaxis.major_label_orientation = "vertical"
    p.xaxis.formatter = NumeralTickFormatter(format='0.0a')  # million
    return p

def chart_histogram(bins_edges: np.ndarray, hists: np.ndarray, 
            quantile_stat: Optional[QuantileStat] = None, desc_stat: Optional[DescStat] = None, 
            size: tuple = (1200, 500), title: str ='Distribution chart') -> Figure:
    """generate a bokeh histogram chart for numeric varaibles

    :param np.ndarray bins_edges: bin edges of histogram (length N + 1)
    :param np.ndarray hists: count of each bin (length N)
    :param QuantileStat quantile_stat: quantile stat object, defaults to None
    :param DescStat desc_stat: descriptive stat object, defaults to None
    :param tuple size: size of the chart, defaults to (500, 400)
    :param str title: title of the chart, defaults to 'Distribution chart'
    :return Figure: bokeh histogram chart
    """
    distr_bins = pd.DataFrame({
        'lower' : bins_edges[:-1],
        'upper' : bins_edges[1:],
        'counts' : hists,
    })
    distr_bins['percent'] = distr_bins['counts'] / distr_bins['counts'].sum()
    max_height = hists.max() * 1.2

    source_hist_cali = ColumnDataSource(distr_bins)
    
    p = figure(plot_width=size[0], plot_height=size[1], title=title,
        x_axis_label='Value', y_axis_label='Counts')
    
    # optional hareas
    if desc_stat is not None:
        p.harea(
            y=[0, max_height], 
            x1 = desc_stat.mean - 3 * desc_stat.std, 
            x2 = desc_stat.mean + 3 * desc_stat.std,
            fill_color="#0687AD", 
            alpha = 0.2, 
            legend_label="± 3σ"
        )
        
    if quantile_stat is not None:
        p.harea(
            y=[0, max_height], 
            x1 = quantile_stat.perc_1th, 
            x2 = quantile_stat.perc_99th,
            fill_color="#06AD1A",
            alpha = 0.2, 
            legend_label="Perc 1-99"
        )

        p.harea(
            y=[0, max_height], 
            x1 = quantile_stat.q3 - 1.5 * quantile_stat.iqr, 
            x2 = quantile_stat.q3 + 1.5 * quantile_stat.iqr,
            fill_color="#E05907", 
            alpha = 0.2, 
            legend_label="IQR"
        )

    # histogram
    p.quad(source=source_hist_cali, bottom=0, top='counts',
        left='lower', right='upper',
        fill_color='#465E83', line_color='black'
    )
        
    # optional vlines
    if quantile_stat is not None:
        p.line(
            x = [quantile_stat.median, quantile_stat.median], 
            y = [0, max_height], 
            legend_label = 'median', 
            line_color='#06AD1A', 
            line_width = 2, 
            line_dash = 'dashed'
        )
    if desc_stat is not None:
        p.line(
            x = [desc_stat.mean, desc_stat.mean], 
            y = [0, max_height], 
            legend_label = 'mean', 
            line_color='#07B2E0', 
            line_width = 2, 
            line_dash = 'dashed'
        )
        
        
    p.y_range = Range1d(start=0, end = max_height)
    p.yaxis.formatter = NumeralTickFormatter(format='0.0a')  # million
    p.add_tools(
        HoverTool(
            tooltips=[
                ("Range", "@lower to @upper"), 
                ("Count", "@counts"), 
                ("Percent", "@percent{(0.00%)}")
            ]
        )
    )
    return p

def chart_categ_distr(vcounts: pd.Series, size: tuple = (1200, 600), 
        title: str ='Distribution chart') -> Figure:
    """generate categ bar chart for vcounts

    :param pd.Series vcounts: the count series (index = categ, value = counts)
    :param tuple size: size of the chart, defaults to (500, 400)
    :param str title: title of the chart, defaults to 'Distribution chart'
    :return Figure: bokeh bar chart
    """
    fig = chart_barchart(
        vcounts,
        max_bar = 20,
        size = size,
        title = title
    )

    return fig


def metric_div(value: float, perc: float, label:str, 
        threshold_p: float = 1.0, back_color: str = 'white') -> Div:
    html = """
    <style>
        h1 {text-align: center;}
        p {text-align: center}
        div {text-align: center;}

        .value {
            text-align: center;
            font-size: 20px;
            color: rgb(44, 144, 129);
            line-height:5px;
        }
        .perc {
            text-align: center;
            font-size: 15px;
            color: rgb(163, 113, 33);
            line-height:5px;
        }
        .label {
            text-align: center;
            font-size: 30px;
            color: rgb(1, 1, 1);
            line-height:25px;
        }
    </style>
    """
    style = "color:red" if perc > threshold_p else ""
    html +=  f"""
    <p class="value">{value:,.0f}</p>
    <p class="perc">({perc:.1%})</p>
    <p class="label" style="{style}">{label}</p>
    """
    return Div(
        text = html, 
        background = back_color, 
        align = 'center', 
        min_width = 200,
        margin = (2, 2, 2, 2),
        width_policy = 'max',
    )

def numeric_count_summary(total: int, missing: StatVar, zeros: StatVar, 
        inf_pos: StatVar, inf_neg: StatVar, xtreme: StatVar) -> layout:
    """numeric variable count part summary chart

    :param int total: total number of samples
    :param StatVar missing: number/percent of missing samples
    :param StatVar zeros: number/percent of zero samples
    :param StatVar inf_pos: number/percent of +inf samples
    :param StatVar inf_neg: number/percent of -inf samples
    :param StatVar xtreme: number/percent of extreme value samples
    :return layout: bokeh layout object
    """
    valid = total - missing.value - zeros.value - inf_pos.value - inf_neg.value - xtreme.value
    valid_perc = 1 - missing.perc - zeros.perc - inf_pos.perc - inf_neg.perc - xtreme.perc

    rows = column([
            row(
                [
                    metric_div(valid, valid_perc, 'Valid', threshold_p = 1, back_color = 'ivory'),
                    metric_div(missing.value, missing.perc, 'Missing', threshold_p = 0.1, back_color = 'gainsboro'),
                    metric_div(zeros.value, zeros.perc, 'Zeros', threshold_p = 0.1, back_color = 'azure'),
                ],  
                sizing_mode='stretch_both'
            ),
            row(
                [
                    metric_div(inf_pos.value, inf_pos.perc, '+Inf', threshold_p = 0.01, back_color = 'lightpink'),
                    metric_div(inf_neg.value, inf_neg.perc, '-Inf', threshold_p = 0.01, back_color = 'lightgreen'),
                    metric_div(xtreme.value, xtreme.perc, 'Xtreme', threshold_p = 0.05, back_color = 'peachpuff')
                ],  
                sizing_mode='stretch_both'
            )
        ],
        sizing_mode='stretch_both'
    )
    return rows

def categ_count_summary(total: int, missing: StatVar, unique: StatVar) -> layout:
    """categorical variable count part summary chart

    :param int total: total number of samples
    :param StatVar missing: number/percent of missing samples
    :param StatVar unique: number/percent of unique samples
    :return layout: bokeh layout object
    """
    valid = total - missing.value
    valid_perc = 1 - missing.perc
    rows = row(
        [
            metric_div(valid, valid_perc, 'Valid', threshold_p = 1, back_color = 'ivory'),
            metric_div(missing.value, missing.perc, 'Missing', threshold_p = 0.1, back_color = 'gainsboro'),
            metric_div(unique.value, unique.perc, 'Unique', back_color = 'aquamarine'),
        ],  
        sizing_mode='stretch_both'
    )
    return rows

def numeric_donut(total: int, missing: StatVar, zeros: StatVar, 
            inf_pos: StatVar, inf_neg: StatVar, xtreme: StatVar,
            size: tuple = (600, 500), title: str = 'Composite Donut') -> Figure:
    """numeric value donut chart

    :param int total: total number of samples
    :param StatVar missing: number/percent of missing samples
    :param StatVar zeros: number/percent of zero samples
    :param StatVar inf_pos: number/percent of +inf samples
    :param StatVar inf_neg: number/percent of -inf samples
    :param StatVar xtreme: number/percent of extreme value samples
    :param tuple size: size of the chart, defaults to (600, 500)
    :param str title: title of the chart, defaults to 'Composite Donut'
    :return Figure: bokeh donut chart
    """
    valid = total - missing.value - zeros.value - inf_pos.value - inf_neg.value - xtreme.value
    donut_dict = pd.Series(
        [valid, missing.value, zeros.value, inf_pos.value, inf_neg.value,  xtreme.value],
        index = ['Valid', 'Missing', 'Zeros', '+Inf', '-Inf', 'Xtreme'],
    )

    fig = chart_donut(
        donut_dict,
        colors = ['gold', 'darkgray', 'azure', 'mediumvioletred', 'olivedrab', 'tomato'],
        size = size,
        title = title
    )
    return fig

def categ_donut(total: int, missing: StatVar, size: tuple = (600, 500), 
                title: str = 'Composite Donut') -> Figure:
    """categ value donut chart

    :param int total: total number of samples
    :param StatVar missing: number/percent of missing samples
    :param tuple size: size of the chart, defaults to (600, 500)
    :param str title: title of the chart, defaults to 'Composite Donut'
    :return Figure: bokeh donut chart
    """
    valid = total - missing.value
    donut_dict = pd.Series(
        [valid, missing.value],
        index = ['Valid', 'Missing'],
    )

    fig = chart_donut(
        donut_dict,
        colors = ['gold', 'darkgray'],
        size = size,
        title = title
    )
    return fig
        


def table_stat(stats: dict) -> DataTable:
    """compile a bokeh data table object

    :param dict stats: the dictionary of data
    :return DataTable: bokeh data table
    """
    df = (
        pd.Series(stats)
        .reset_index(name='VALUE')
        .rename(columns = {'index' : 'METRIC'})
    )
    source = ColumnDataSource(df)
    data_table = DataTable(
        source=source, 
        columns=[
            TableColumn(
                field = 'METRIC',
                title = 'METRIC',
                sortable = False,
                formatter = StringFormatter(
                    font_style = 'bold'
                )
            ),
            TableColumn(
                field = 'VALUE',
                title = 'VALUE',
                sortable = False,
                formatter = ScientificFormatter(
                    precision = 5,
                    text_color = 'darkslategrey',
                )
            )
        ], 
        editable=False,
        index_position = None,
        height = None,
        sizing_mode = 'stretch_both'
    )
    return data_table

def chart_boxplot(p_x_y: pd.DataFrame, xname:str = 'x', yname:str = 'y', 
        size: tuple = (500, 400), title: str='Boxplot') -> Figure:
    """box plot for conditional probability

    :param pd.DataFrame p_x_y: the conditional prob(x|y)
    :param str xname: independent variable name, defaults to 'x'
    :param str yname: dependent variable name, defaults to 'y'
    :param tuple size: size of the chart, defaults to (500, 400)
    :param str title: title of the chart, defaults to 'Boxplot'
    :return Figure: bokeh boxplot chart
    """
    p_x_y['y'] = p_x_y.index.astype(str)
    data_source = ColumnDataSource(p_x_y)

    p = figure(plot_width=size[0], plot_height=size[1],
            title=title, x_range=data_source.data['y'],
            x_axis_label=yname, y_axis_label=xname)


    # boxes
    p.vbar(x='y', width=0.7, bottom='median', top='q3', 
        source=data_source, fill_color="#E08E79", line_color="black")
    p.vbar(x='y', width=0.7, bottom='q1', top='median', 
        source=data_source, fill_color="#3B8686", line_color="black")
    # whiskers (almost-0 height rects simpler than segments)
    p.rect(x='y', y='rbound', width = 0.2, height = 0.005, 
        source=data_source, line_color="black")
    p.rect(x='y', y='lbound', width = 0.2, height = 0.005, 
        source=data_source, line_color="black")
    p.rect(x='y', y='mean', width = 0.7, height = 0.005, 
        source=data_source, line_color="#B63B3B", fill_color="#B63B3B",
        legend_label='mean')
    # stems
    p.segment(x0='y', y0='rbound', x1='y', y1='q3', 
        source=data_source, line_color="black",
    )
    p.segment(x0='y', y0='lbound', x1='y', y1='q1', 
        source=data_source, line_color="black",
    )

    p.add_tools(HoverTool(tooltips=[
        ("label", "@y"), 
        ("mean", "@mean"), 
        ("lbound", "@lbound"), 
        ("q1", "@q1"),
        ("median", "@median"), 
        ("q3", "@q3"), 
        ("rbound", "@rbound")
    ]))

    p.xgrid.grid_line_color = None

    #p.yaxis.major_label_orientation = "vertical"
    p.yaxis.formatter = NumeralTickFormatter(format='0.0a')  # million
    
    return p

def chart_optbin_multiclass(bin_r: pd.DataFrame, 
        ylabels: List[str] = None, xname:str = 'x', yname:str = 'y',
        size: tuple = (1000, 500), title: str = "Realized Event vs. Bins") -> Figure:
    """_sumOptimal binning result chartmary_

    :param pd.DataFrame bin_r: results returned by optbin.binning_table.build(add_totals=False)
                    columns [Bin, Count, Count (%), Event_x, Event rate_y]
    :param List[str] ylabels: if given, use this list to identify the columns, defaults to None
    :param str xname: independent variable name, defaults to 'x'
    :param str yname: dependent variable name, defaults to 'y'
    :param tuple size: size of the chart, defaults to (1000, 500)
    :param str title: title of the chart, defaults to 'Realized Event vs. Bins'
    :return Figure: bokeh stack bar chart
    """

    # extract target labels
    if ylabels is None:
        ylabels =(
            bin_r.columns
            [bin_r.columns.str.startswith('Event_rate_')]
            .str.replace('Event_rate_', '')
            .tolist()
        )
    
    color = cycle(Dark2_8)
    colors = [next(color) for _ in ylabels]
    
    source_bin = ColumnDataSource(bin_r)

    fig_bin = figure(plot_width=size[0], plot_height=size[1],
        x_axis_label = xname, y_axis_label = yname,
        x_range=source_bin.data['Bin'], title = title)

    fig_bin.add_layout(Legend(), 'right')
    fig_bin.vbar_stack(
        [f"Event_{y_label}" for y_label in ylabels], # ['Event_0', 'Event_1'] 
        legend_label = ylabels,
        source = source_bin, x = 'Bin', width = 0.8,
        line_color = 'black', 
        color = colors
    )

    fig_bin.yaxis.formatter = NumeralTickFormatter(format='0.0a')  # million

    # add event rate to secondary axis
    #fig_bin.y_range = Range1d(start = 0, end = 1)
    fig_bin.extra_y_ranges = {"EVENT_RT_AXIS": Range1d(start=0, end = 1.2)}
    fig_bin.add_layout(LinearAxis(y_range_name="EVENT_RT_AXIS"), 'right')
    
    for i, ylabel in enumerate(ylabels):
        fig_bin.line(x='Bin', y=f'Event_rate_{ylabel}', source=source_bin, 
            y_range_name='EVENT_RT_AXIS', width=3,
            legend_label=f'Event_rate_{ylabel}', color=colors[i]
        )
        fig_bin.circle(x='Bin', y=f'Event_rate_{ylabel}', source=source_bin, 
            y_range_name='EVENT_RT_AXIS', color='black',
            fill_color="white", size=10
        )
    
    fig_bin.add_tools(
        HoverTool(
            tooltips=[
                ("Bin", "@Bin"), 
                ("Count", "@Count"), 
                ("Count (%)", "@{Count (%)}{(0.000)}")
            ] 
            + [(ylabel, "@{Event_" + ylabel + "}") 
               for ylabel in ylabels] 
            + [(f"Event_rate_{ylabel}", "@{Event_rate_" + ylabel + "}{(0.0%)}") 
               for ylabel in ylabels]
    ))
    fig_bin.xaxis.major_label_orientation = "vertical"
    fig_bin.y_range = Range1d(
        start = 0, 
        end = bin_r['Count'].max() * 1.2
    )

    return fig_bin

def chart_segment_group_count(seg_df: Dict[str, pd.DataFrame], 
        group_name:str = 'y',  agg_name:str = 'x', size: tuple = (500, 500), 
        title: str = "Stacking Segment Count by Group") -> Figure:
    """stacking segment count by group

    :param Dict[str, pd.DataFrame] seg_df: usually the p_x_y, distribution of x given y
    :param str group_name: group column name, defaults to 'y'
    :param str agg_name: aggregation on column name, defaults to 'x'
    :param tuple size: size of the chart, defaults to (500, 500)
    :param str title: title of the chart, defaults to 'Stacking Segment Count by Group'
    :return Figure: bokeh stack bar chart
    """
    seg = pd.concat(
        [s['count'].rename(group) for group, s in seg_df.items()], 
        axis = 1
    ).T
    seg.columns = seg.columns.astype('str')
    aggs = seg.columns.astype('str').tolist()
    seg[group_name] = seg.index.astype(str)
    
    color = cycle(Dark2_8)
    colors = [next(color) for _ in aggs]
    
    source_bin = ColumnDataSource(seg)

    p = figure(plot_width=size[0], plot_height=size[1],
        x_axis_label = group_name, y_axis_label = agg_name,
        x_range=source_bin.data[group_name], title = title)

    p.add_layout(Legend(), 'right')
    p.vbar_stack(
        aggs,
        x = group_name,
        source = source_bin,
        legend_label = aggs,
        width = 0.8,
        line_color = 'black', 
        color = colors
    )

    p.yaxis.formatter = NumeralTickFormatter(format='0.0a')  # million
    
    p.add_tools(
        HoverTool(
            tooltips=[
                ("group", f"@{group_name}")
            ]
            + [(agglabel, "@{" + agglabel + "}") 
               for agglabel in aggs] 
    ))
    
    return p

def chart_segment_group_perc(seg_df: Dict[str, pd.DataFrame], 
        group_name:str = 'y', agg_name:str = 'x', size: tuple = (500, 500), 
        title: str = "Segment Distribution by Group") -> Figure:
    """stacking segment count percentage by group

    :param Dict[str, pd.DataFrame] seg_df: usually the p_x_y, distribution of x given y
    :param str group_name: group column name, defaults to 'y'
    :param str agg_name: aggregation on column name, defaults to 'x'
    :param tuple size: size of the chart, defaults to (500, 500)
    :param str title: title of the chart, defaults to 'Stacking Distribution by Group'
    :return Figure: bokeh stack bar chart
    """
    seg = pd.concat(
        [s['perc'].rename(group) for group, s in seg_df.items()], 
        axis = 1
    ).T
    seg.columns = seg.columns.astype('str')
    aggs = seg.columns.astype('str').tolist()
    seg[group_name] = seg.index.astype(str)
    
    color = cycle(Dark2_8)
    colors = [next(color) for _ in aggs]
    
    source_bin = ColumnDataSource(seg)

    p = figure(plot_width=size[0], plot_height=size[1],
        x_axis_label = group_name, y_axis_label = agg_name,
        x_range=source_bin.data[group_name], title = title)

    p.add_layout(Legend(), 'right')
    
    for i, agg in enumerate(aggs):
        p.line(x=group_name, y=agg, source=source_bin, 
            width=3, legend_label=agg, color=colors[i]
        )
        p.circle(x=group_name, y=agg, source=source_bin, 
            color='black', fill_color="white", size=10
        )

    p.yaxis.formatter = NumeralTickFormatter(format='0.00%')  # million
    
    p.add_tools(
        HoverTool(
            tooltips=[
                ("group", f"@{group_name}")
            ]
            + [(agglabel, "@{" + agglabel +"}{(0.00%)}") 
               for agglabel in aggs] 
    ))
    
    return p

def plot_numeric(nst: NumericStatObj, html_path: Optional[str] = None) -> column:
    """plot numeric variable EDA result

    :param NumericStatObj nst: the stat obj of numeric variable
    :param Optional[str] html_path: if given, will output to local file
    :return Figure: bokeh layout object
    """
    summary = numeric_count_summary(
        nst.summary.total_, 
        nst.summary.missing_, 
        nst.summary.zeros_, 
        nst.summary.infs_pos_, 
        nst.summary.infs_neg_, 
        nst.summary.xtreme_
    )
    donut = numeric_donut(
        nst.summary.total_, 
        nst.summary.missing_, 
        nst.summary.zeros_, 
        nst.summary.infs_pos_, 
        nst.summary.infs_neg_, 
        nst.summary.xtreme_,
        size = (600, 500)
    )

    tabs_desc_stat = [Panel(
        child = table_stat(asdict(nst.summary.stat_descriptive_)),
        title="Origin"
    )]
    tabs_quant_stat = [Panel(
        child = table_stat(asdict(nst.summary.stat_quantile_)),
        title="Origin"
    )]
    if nst.attr.log_scale_:
        tabs_desc_stat.append(Panel(
            child = table_stat(asdict(nst.summary.stat_descriptive_log_)),
            title="Log"
        ))
        tabs_quant_stat.append(Panel(
            child = table_stat(asdict(nst.summary.stat_quantile_log_)),
            title="Log"
        ))

    tabs_hist = [Panel(
        child = chart_histogram(
            bins_edges = nst.summary.bin_edges_,
            hists = nst.summary.hist_,
            quantile_stat=nst.summary.stat_quantile_,
            desc_stat=nst.summary.stat_descriptive_,
            title = 'Histogram for Valid Values',
            size = (1200, 500)
        ),
        title="Histogram - Origin"
    )]
    if nst.attr.log_scale_:
        tabs_hist.append(Panel(
            child = chart_histogram(
                bins_edges = nst.summary.bin_edges_log_,
                hists = nst.summary.hist_log_,
                quantile_stat=nst.summary.stat_quantile_log_,
                desc_stat=nst.summary.stat_descriptive_log_,
                title = 'Histogram for Valid Values',
                size = (1200, 500)
            ),
            title="Statistics - Log"
        ))

    # assemble to widget
    table_widgets = [
        Div(text="""
            <h3 style="text-align:left">Quantile Statistics for Valid Values</h2>
        """),
        Tabs(tabs=tabs_quant_stat),
        Div(text="""
            <h3 style="text-align:left">Descriptive Statistics for Valid Values</h2>
        """),
        Tabs(tabs=tabs_desc_stat),
    ]

    if nst.attr.xtreme_method_ is not None:
        tabs_xtreme = [Panel(
            child = table_stat(asdict(nst.summary.xtreme_stat_)),
            title="Origin"
        )]
        if nst.attr.log_scale_:
            tabs_xtreme.append(Panel(
                child = table_stat(asdict(nst.summary.xtreme_stat_log_)),
                title="Log"
            ))
        table_widgets.extend([
            Div(text=f"""
                <h3>Xtreme Value Statistics</h2><p style="text-align:left">method = {nst.attr.xtreme_method_}</p>
            """),
            Tabs(tabs=tabs_xtreme),
        ])

    result = column([
        row([
            column([
                summary, 
                donut,
            ],
            sizing_mode = 'stretch_width'
            ),
            column(
                table_widgets,
                sizing_mode = 'stretch_width'
            )
        ]),
        Tabs(tabs=tabs_hist),
    ],
    #sizing_mode = 'fixed'
    )
    
    if html_path:
        output_file(html_path)
        save(result)
        
    return result


def plot_categ(nst: CategStatObj, html_path: Optional[str] = None) -> row:
    """plot categorical variable EDA result (generic)

    :param CategStatObj nst: the stat obj of categorical variable
    :param Optional[str] html_path:if given, will output to local file
    :return Figure: bokeh layout object
    """
    summary = categ_count_summary(
        nst.summary.total_, 
        nst.summary.missing_, 
        nst.summary.unique_
    )
    donut = categ_donut(
        total=nst.summary.total_,
        missing=nst.summary.missing_,
        size = (600, 500)
    )
    distr = chart_categ_distr(
        vcounts = nst.summary.vcounts_,
        size = (600, 640)
    )

    result = row([
        column([
            summary,
            donut
        ],
        sizing_mode = 'stretch_width'
        ),
        distr
    ],
    sizing_mode = 'stretch_width'
    )
    
    if html_path:
        output_file(html_path)
        save(result)
        
    return result

def plot_binary(nst: BinaryStatObj, html_path: Optional[str] = None) -> row:
    """"plot binary categ variable EDA result

    :param BinaryStatObj nst: the stat obj of binary categorical variable
    :param Optional[str] html_path: if given, will output to local file
    :return Figure: bokeh layout object
    """
    summary = categ_count_summary(
        nst.summary.total_, 
        nst.summary.missing_, 
        nst.summary.unique_
    )
    donut = categ_donut(
        total=nst.summary.total_,
        missing=nst.summary.missing_,
        size = (600, 500)
    )
    distr = chart_categ_distr(
        vcounts = nst.summary.binary_vcounts_,
        size = (600, 640)
    )

    result = row([
        column([
            summary,
            donut
        ],
        sizing_mode = 'stretch_width'
        ),
        distr
    ],
    sizing_mode = 'stretch_width'
    )
    
    if html_path:
        output_file(html_path)
        save(result)
        
    return result

def plot_numeric_clf_target_corr(nuct: NumericUniVarClfTargetCorr, 
        html_path: str = None) -> row:
    """plot univariate numeric feature correlation with classifier target summary

    :param NumericUniVarClfTargetCorr nuct: univariate numeric feature correlation with classifier target
    :param str html_path: if given, will output to local file
    :return Figure: bokeh layout object
    """
    tabs_x_y = Tabs(
        tabs = [
            Panel(
                child = chart_boxplot(
                    nuct.p_x_y_['origin'],
                    xname=nuct.colname_, 
                    yname=nuct.yname_,
                    size=(400, 500),
                    title = "P(x|y) - Feature Distribution By Target"
                ),
                title="Origin"
            ),
            Panel(
                child = chart_boxplot(
                    nuct.p_x_y_['log'],
                    xname=nuct.colname_, 
                    yname=nuct.yname_,
                    size=(400, 500),
                    title = "P(x|y) - Feature Distribution By Target"
                ),
                title="Log"
            )
        ]
    )
    
    fig_y_x = chart_optbin_multiclass(
        nuct.p_y_x_,
        ylabels = nuct.ylabels_,
        xname=nuct.colname_, 
        yname=nuct.yname_,
        size=(1000, 500),
        title = "P(y|x) - Event Rate by Bucketized Feature"
    )
    
    
    result = row([fig_y_x, tabs_x_y], 
        sizing_mode='stretch_height')
    
    if html_path:
        output_file(html_path)
        save(result)
    return result
    
    
def plot_categ_clf_target_corr(cuct: CategUniVarClfTargetCorr, 
        html_path: str = None) -> row:
    """plot univariate categ feature correlation with classifier target summary

    :param CategUniVarClfTargetCorr cuct: if given, will output to local file
    :return Figure: bokeh layout object
    :return row: _description_
    """
    tabs_x_y = Tabs(
        tabs = [
            Panel(
                child = chart_segment_group_count(
                    cuct.p_x_y_,
                    group_name=cuct.yname_, 
                    agg_name=cuct.colname_,
                    size=(600, 500),
                    title = "P(x|y) - Category Count By Target"
                ),
                title="Count"
            ),
            Panel(
                child = chart_segment_group_perc(
                    cuct.p_x_y_,
                    group_name=cuct.yname_, 
                    agg_name=cuct.colname_,
                    size=(600, 500),
                    title = "P(x|y) - Category Distribution By Target"
                ),
                title="Percentage"
            )
        ]
    )
    
    tabs_y_x = Tabs(
        tabs = [
            Panel(
                child = chart_segment_group_count(
                    cuct.p_y_x_,
                    group_name=cuct.colname_, 
                    agg_name=cuct.yname_,
                    size=(800, 500),
                    title = "P(y|x) - Event Count by Feature Categories"
                ),
                title="Count"
            ),
            Panel(
                child = chart_segment_group_perc(
                    cuct.p_y_x_,
                    group_name=cuct.colname_, 
                    agg_name=cuct.yname_,
                    size=(800, 500),
                    title = "P(x|y) - Event Rate by Feature Categories"
                ),
                title="Percentage"
            )
        ]
    )
    result = row([tabs_y_x, tabs_x_y], 
        sizing_mode='stretch_height')
    
    if html_path:
        output_file(html_path)
        save(result)
    return result

def table_summary(ts: TabularStat) -> row:
    """get tabular feature count summary

    :param TabularStat ts: tabular stat object
    :return row: bokeh row layout
    """
    num_numeric = len(ts.get_numeric_cols())
    num_binary = len(ts.get_binary_cols())
    num_ordinal = len(ts.get_ordinal_cols())
    num_nominal = len(ts.get_nominal_cols())
    total = len(ts)
    rows = row(
        [
            metric_div(
                value = num_numeric, 
                perc = num_numeric / total, 
                label = 'Numerical Features', 
                threshold_p = 1
            ),
            metric_div(
                value = num_binary, 
                perc = num_binary / total, 
                label = 'Binary Features', 
                threshold_p = 1
            ),
            metric_div(
                value = num_ordinal, 
                perc = num_ordinal / total, 
                label = 'Ordinal Categorical Features', 
                threshold_p = 1
            ),
            metric_div(
                value = num_nominal, 
                perc = num_nominal / total, 
                label = 'Nominal Categorical Features', 
                threshold_p = 1
            ),
        ],  
        sizing_mode='stretch_both'
    )
    return rows

def plot_table_profiling(ts: TabularStat, html_path: Optional[str] = None) -> column:
    """get tabular feature profiling plot

    :param TabularStat ts: tabular stat object
    :param str html_path: if given, will output to local file
    :return row: bokeh row layout
    """
    figs = [
        Div(
            text = """
                <h1>Overview</h1>
            """
        ),
        table_summary(ts)
    ]
    cols_numeric = ts.get_numeric_cols()
    cols_binary = ts.get_binary_cols()
    cols_ordinal = ts.get_ordinal_cols()
    cols_nominal = ts.get_nominal_cols()
    for col, config in ts.items():
        if col in cols_numeric:
            fig = plot_numeric(config)
        elif col in cols_ordinal + cols_nominal:
            fig = plot_categ(config)
        elif col in cols_binary:
            fig = plot_binary(config)

        figs.append(
            Div(
                text = f"""
                    <h2 style="color:#395687">{col}</h2>
                """
            )
        )
        figs.append(fig)
    
    result = column(figs, sizing_mode = 'stretch_both')
    
    if html_path:
        output_file(html_path)
        save(result)
        
    return result

def plot_uni_clf_target_corr(tuvct: Dict[str, BaseUniVarClfTargetCorr], 
            html_path: Optional[str] = None) -> column:
    """get univariate features correlation with classifier target plots

    :param Dict[str, BaseUniVarClfTargetCorr] tuvct: dict of corr summary
    :param str html_path: if given, will output to local file
    :return row: bokeh column layout
    """
    figs = [
        Div(
            text = """
                <h1>Univaraite Feature-Target Correlation for Classification</h1>
            """
        )
    ]
    for col, config in tuvct.items():
        if isinstance(config, NumericUniVarClfTargetCorr):
            fig = plot_numeric_clf_target_corr(config)
        elif isinstance(config, CategUniVarClfTargetCorr):
            fig = plot_categ_clf_target_corr(config)

        figs.append(
            Div(
                text = f"""
                    <h2 style="color:#395687">{col}</h2>
                """
            )
        )
        figs.append(fig)
    
    result = column(figs, sizing_mode = 'stretch_both')
    
    if html_path:
        output_file(html_path)
        save(result)
        
    return result


####### Other plots

def chart_gridplot(grid: pd.DataFrame, reverse_color:bool = False, 
        size: tuple = (1000, 600), title: str ='Grid Plot') -> layout:
    """grid plot

    :param pd.DataFrame grid: underlying dataframe
    :param bool reverse_color: whether to reverse color order, defaults to False
    :param tuple size: size of the chart, defaults to (1000, 600)
    :param str title: title of the chart, defaults to 'Grid Plot'
    :return layout: bokeh layout object
    """
    grid_flat = (
        grid
        .stack()
        .rename("value")
        .reset_index()
    )

    grid_flat_source = ColumnDataSource(grid_flat)
    mapper_color = LinearColorMapper(
        palette=Magma256[::-1] if reverse_color else Magma256
    )
    p = figure(
        plot_width=size[0], plot_height=size[1], title=title,
        x_axis_label=grid.columns.name, 
        y_axis_label=grid.index.name,
        x_range = grid.columns.tolist(), 
        y_range = grid.index.tolist()
    )
    p.rect(
        x=grid.columns.name,
        y=grid.index.name,
        width=1,
        height=1,
        source=grid_flat_source,
        line_color=None,
        fill_color=transform('value', mapper_color)
    )

    p.line(x = grid.index.tolist(), y = grid.columns.tolist(), 
           color='red', width=2, line_dash="4 4")

    # Add legend
    color_bar = ColorBar(
        color_mapper=mapper_color,
        location=(0, 0),
        ticker=BasicTicker(desired_num_ticks=256)
    )

    p.xaxis.major_tick_line_color = None  # turn off x-axis major ticks
    p.xaxis.minor_tick_line_color = None  # turn off x-axis minor ticks
    p.yaxis.major_tick_line_color = None  # turn off y-axis major ticks
    p.yaxis.minor_tick_line_color = None  # turn off y-axis minor ticks
    p.xaxis.major_label_orientation = "vertical"
    p.add_layout(color_bar, 'right')
    p.add_tools(
        HoverTool(
            tooltips=[
                (grid.index.name, f"@{grid.index.name}"), 
                (grid.columns.name, f"@{grid.columns.name}"),
                ("value", "@value")
            ]
        )
    )

    return p


import pymongo
from pymongo.mongo_client import MongoClient
from luntaiDs.CommonTools.schema_manager import BaseSchemaManager


class MongoSchemaManager(BaseSchemaManager):
    
    def __init__(self, mongo_client: MongoClient, database: str, collection: str):
        """schema manager for mongo db
        
        the mongodb implementation works like this, under given database/collection
        mongodb
            database
                collection
                    entry1: schema A, table 1, config
                    entry2: schema A, table 2, config
                    entry3: schema B, table 1, config
                    ...

        :param MongoClient mongo_client: mongo db python connector object
        :param str database: the mongo db database name to save the table config
        :param str collection: the mongo db table name to save the table config
        """
        self._mongo_client = mongo_client
        self.database = database
        self.collection = collection
    
    def write_raw(self, schema: str, table: str, content: dict):
        """handle how to write raw (dictionary) config into given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :param dict content: the dict version of Dschema object
        """
        # add more keys
        content = {
            'schema': schema, 
            'table': table,
            'columns' : content,
        }
        db = self._mongo_client[self.database]
        collection = db[self.collection]
        collection.replace_one(
            filter = {
                'schema': schema,
                'table': table
            }, # find matching record, if any
            replacement = content,
            upsert = True # update if found, insert if not found
        )
    
    def read_raw(self, schema: str, table: str) -> dict:
        """handle how to write read (dictionary) config from given schema/table

        :param str schema: the schema name
        :param str table: the table name
        :return dict: the dict version of Dschema object 
        """
        db = self._mongo_client[self.database]
        collection = db[self.collection]
        record = (
            collection
            .find_one(
                {'schema': schema,'table': table,}, # matching condition
                {'_id': 0,}, # drop id column
                sort = [( '_id', pymongo.DESCENDING )] # in case multiple, find the latest record
            )
        )
        if record is None:
            raise ValueError("No record found for given schema and table")
        return record.get('columns') # only need the columns field


import json
import requests
from requests.auth import HTTPBasicAuth
from typing import List, Dict, Any

class AirflowAPI:
    # https://airflow.apache.org/docs/apache-airflow/stable/security/api.html#api-authentication
    def __init__(self, host: str, port: int, username: str, password: str) -> None:
        self._base_url = f"http://{host}:{port}/api/v1"
        self._auth = HTTPBasicAuth(username, password)
        
    def get_variable_list(self) -> List[str]:
        r = requests.get(
            url = f"{self._base_url}/variables",
            auth = self._auth
        )
        if r.status_code != 200:
            r.raise_for_status()
        r = r.json().get('variables')
        return [i.get("key") for i in r]
    
    def get_variable(self, key: str) -> Any:
        r = requests.get(
            url = f"{self._base_url}/variables/{key}",
            auth = self._auth
        )
        if r.status_code != 200:
            r.raise_for_status()
        return r.json().get('value')
    
    def upsert_variable(self, key: str, value: Any, descr: str = ""):
        # see if in variable already
        vars = self.get_variable_list()
        if key in vars:
            # update
            r = requests.patch(
                url = f"{self._base_url}/variables/{key}",
                auth = self._auth,
                json = {
                    "description": descr,
                    "key": key,
                    "value": json.dumps(value)
                }
            )
        else:
            # insert
            r = requests.post(
                url = f"{self._base_url}/variables",
                auth = self._auth,
                json = {
                    "description": descr,
                    "key": key,
                    "value": json.dumps(value)
                }
            )
        if r.status_code != 200:
            r.raise_for_status()
            
    def delete_variable(self, key: str):
        r = requests.delete(
            url = f"{self._base_url}/variables/{key}",
            auth = self._auth
        )
        if r.status_code != 200:
            r.raise_for_status()
    


import json
import logging
import os
from collections import OrderedDict



''' ---- 			file loader 			---- '''

def loadJSON(file):
    if isinstance(file, (list, dict)):
        return file
    elif isinstance(file, str):
        with open(file, "rb") as obj:
            return json.load(obj, object_pairs_hook = OrderedDict)

    else:
        raise ValueError("Please parse a file path or JS object")


def toJSON(js, file):
    with open(file, "w") as obj:
        json.dump(js, obj, indent = 4)
